from typing import NamedTuple

class VersionInfo(NamedTuple):
    major: int
    minor: int
    patch: int

class BuildInfo(NamedTuple):
    version: str
    version_info: VersionInfo
    so_version: str
    full_so_version: str
    compiler_id: str
    compiler_version: str
    compiler_flags: str
    git_id: str
    git_description: str
    package_kind: str
    build_type: str

class RuntimeInfo(NamedTuple):
    simd_level: str
    detected_simd_level: str

cpp_build_info: BuildInfo
cpp_version: str
cpp_version_info: VersionInfo

def runtime_info() -> RuntimeInfo: ...
def set_timezone_db_path(path: str) -> None: ...

__all__ = [
    "VersionInfo",
    "BuildInfo",
    "RuntimeInfo",
    "cpp_build_info",
    "cpp_version",
    "cpp_version_info",
    "runtime_info",
    "set_timezone_db_path",
]



from typing import Callable, Iterable, Literal, Sequence, TypeAlias, overload

from _typeshed import StrPath
from pyarrow._dataset import (
    CsvFileFormat,
    CsvFragmentScanOptions,
    Dataset,
    DatasetFactory,
    DirectoryPartitioning,
    FeatherFileFormat,
    FileFormat,
    FileFragment,
    FilenamePartitioning,
    FileSystemDataset,
    FileSystemDatasetFactory,
    FileSystemFactoryOptions,
    FileWriteOptions,
    Fragment,
    FragmentScanOptions,
    HivePartitioning,
    InMemoryDataset,
    IpcFileFormat,
    IpcFileWriteOptions,
    JsonFileFormat,
    JsonFragmentScanOptions,
    Partitioning,
    PartitioningFactory,
    Scanner,
    TaggedRecordBatch,
    UnionDataset,
    UnionDatasetFactory,
    WrittenFile,
    get_partition_keys,
)
from pyarrow._dataset_orc import OrcFileFormat
from pyarrow._dataset_parquet import (
    ParquetDatasetFactory,
    ParquetFactoryOptions,
    ParquetFileFormat,
    ParquetFileFragment,
    ParquetFileWriteOptions,
    ParquetFragmentScanOptions,
    ParquetReadOptions,
    RowGroupInfo,
)
from pyarrow._dataset_parquet_encryption import (
    ParquetDecryptionConfig,
    ParquetEncryptionConfig,
)
from pyarrow.compute import Expression, field, scalar
from pyarrow.lib import Array, RecordBatch, RecordBatchReader, Schema, Table

from ._fs import SupportedFileSystem

_orc_available: bool
_parquet_available: bool

__all__ = [
    "CsvFileFormat",
    "CsvFragmentScanOptions",
    "Dataset",
    "DatasetFactory",
    "DirectoryPartitioning",
    "FeatherFileFormat",
    "FileFormat",
    "FileFragment",
    "FilenamePartitioning",
    "FileSystemDataset",
    "FileSystemDatasetFactory",
    "FileSystemFactoryOptions",
    "FileWriteOptions",
    "Fragment",
    "FragmentScanOptions",
    "HivePartitioning",
    "InMemoryDataset",
    "IpcFileFormat",
    "IpcFileWriteOptions",
    "JsonFileFormat",
    "JsonFragmentScanOptions",
    "Partitioning",
    "PartitioningFactory",
    "Scanner",
    "TaggedRecordBatch",
    "UnionDataset",
    "UnionDatasetFactory",
    "WrittenFile",
    "get_partition_keys",
    # Orc
    "OrcFileFormat",
    # Parquet
    "ParquetDatasetFactory",
    "ParquetFactoryOptions",
    "ParquetFileFormat",
    "ParquetFileFragment",
    "ParquetFileWriteOptions",
    "ParquetFragmentScanOptions",
    "ParquetReadOptions",
    "RowGroupInfo",
    # Parquet Encryption
    "ParquetDecryptionConfig",
    "ParquetEncryptionConfig",
    # Compute
    "Expression",
    "field",
    "scalar",
    # Dataset
    "partitioning",
    "parquet_dataset",
    "write_dataset",
]

_DatasetFormat: TypeAlias = Literal["parquet", "ipc", "arrow", "feather", "csv"]

@overload
def partitioning(
    schema: Schema,
) -> Partitioning: ...
@overload
def partitioning(
    schema: Schema,
    *,
    flavor: Literal["filename"],
    dictionaries: dict[str, Array] | None = None,
) -> Partitioning: ...
@overload
def partitioning(
    schema: Schema,
    *,
    flavor: Literal["filename"],
    dictionaries: Literal["infer"],
) -> PartitioningFactory: ...
@overload
def partitioning(
    field_names: list[str],
    *,
    flavor: Literal["filename"],
) -> PartitioningFactory: ...
@overload
def partitioning(
    schema: Schema,
    *,
    flavor: Literal["hive"],
    dictionaries: Literal["infer"],
) -> PartitioningFactory: ...
@overload
def partitioning(
    *,
    flavor: Literal["hive"],
) -> PartitioningFactory: ...
@overload
def partitioning(
    schema: Schema,
    *,
    flavor: Literal["hive"],
    dictionaries: dict[str, Array] | None = None,
) -> Partitioning: ...
def parquet_dataset(
    metadata_path: StrPath,
    schema: Schema | None = None,
    filesystem: SupportedFileSystem | None = None,
    format: ParquetFileFormat | None = None,
    partitioning: Partitioning | PartitioningFactory | None = None,
    partition_base_dir: str | None = None,
) -> FileSystemDataset: ...
@overload
def dataset(
    source: StrPath | Sequence[StrPath],
    schema: Schema | None = None,
    format: FileFormat | _DatasetFormat | None = None,
    filesystem: SupportedFileSystem | str | None = None,
    partitioning: Partitioning | PartitioningFactory | str | list[str] | None = None,
    partition_base_dir: str | None = None,
    exclude_invalid_files: bool | None = None,
    ignore_prefixes: list[str] | None = None,
) -> FileSystemDataset: ...
@overload
def dataset(
    source: list[Dataset],
    schema: Schema | None = None,
    format: FileFormat | _DatasetFormat | None = None,
    filesystem: SupportedFileSystem | str | None = None,
    partitioning: Partitioning | PartitioningFactory | str | list[str] | None = None,
    partition_base_dir: str | None = None,
    exclude_invalid_files: bool | None = None,
    ignore_prefixes: list[str] | None = None,
) -> UnionDataset: ...
@overload
def dataset(
    source: Iterable[RecordBatch] | Iterable[Table] | RecordBatchReader,
    schema: Schema | None = None,
    format: FileFormat | _DatasetFormat | None = None,
    filesystem: SupportedFileSystem | str | None = None,
    partitioning: Partitioning | PartitioningFactory | str | list[str] | None = None,
    partition_base_dir: str | None = None,
    exclude_invalid_files: bool | None = None,
    ignore_prefixes: list[str] | None = None,
) -> InMemoryDataset: ...
def write_dataset(
    data: Dataset | Table | RecordBatch | RecordBatchReader | list[Table] | Iterable[RecordBatch],
    base_dir: StrPath,
    *,
    basename_template: str | None = None,
    format: FileFormat | _DatasetFormat | None = None,
    partitioning: Partitioning | list[str] | None = None,
    partitioning_flavor: str | None = None,
    schema: Schema | None = None,
    filesystem: SupportedFileSystem | None = None,
    file_options: FileWriteOptions | None = None,
    use_threads: bool = True,
    max_partitions: int = 1024,
    max_open_files: int = 1024,
    max_rows_per_file: int = 0,
    min_rows_per_group: int = 0,
    max_rows_per_group: int = 1024 * 1024,
    file_visitor: Callable[[str], None] | None = None,
    existing_data_behavior: Literal["error", "overwrite_or_ignore", "delete_matching"] = "error",
    create_dir: bool = True,
): ...



import datetime as dt
import enum
import sys

from abc import ABC, abstractmethod

if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self
if sys.version_info >= (3, 10):
    from typing import TypeAlias
else:
    from typing_extensions import TypeAlias

from typing import Union, overload

from fsspec import AbstractFileSystem

from .lib import NativeFile, _Weakrefable

SupportedFileSystem: TypeAlias = Union[AbstractFileSystem, FileSystem]

class FileType(enum.IntFlag):
    NotFound = enum.auto()
    Unknown = enum.auto()
    File = enum.auto()
    Directory = enum.auto()

class FileInfo(_Weakrefable):
    def __init__(
        self,
        path: str,
        type: FileType = FileType.Unknown,
        *,
        mtime: dt.datetime | float | None = None,
        mtime_ns: int | None = None,
        size: int | None = None,
    ): ...
    @property
    def type(self) -> FileType: ...
    @property
    def is_file(self) -> bool: ...
    @property
    def path(self) -> str: ...
    @property
    def base_name(self) -> str: ...
    @property
    def size(self) -> int: ...
    @property
    def extension(self) -> str: ...
    @property
    def mtime(self) -> dt.datetime | None: ...
    @property
    def mtime_ns(self) -> int | None: ...

class FileSelector(_Weakrefable):
    base_dir: str
    allow_not_found: bool
    recursive: bool
    def __init__(self, base_dir: str, allow_not_found: bool = False, recursive: bool = False): ...

class FileSystem(_Weakrefable):
    @classmethod
    def from_uri(cls, uri: str) -> tuple[Self, str]: ...
    def equals(self, other: FileSystem) -> bool: ...
    @property
    def type_name(self) -> str: ...
    @overload
    def get_file_info(self, paths_or_selector: str) -> FileInfo: ...
    @overload
    def get_file_info(self, paths_or_selector: FileSelector | list[str]) -> list[FileInfo]: ...
    def create_dir(self, path: str, *, recursive: bool = True) -> None: ...
    def delete_dir(self, path: str) -> None: ...
    def delete_dir_contents(
        self, path: str, *, accept_root_dir: bool = False, missing_dir_ok: bool = False
    ) -> None: ...
    def move(self, src: str, dest: str) -> None: ...
    def copy_file(self, src: str, dest: str) -> None: ...
    def delete_file(self, path: str) -> None: ...
    def open_input_file(self, path: str) -> NativeFile: ...
    def open_input_stream(
        self, path: str, compression: str | None = "detect", buffer_size: int | None = None
    ) -> NativeFile: ...
    def open_output_stream(
        self,
        path: str,
        compression: str | None = "detect",
        buffer_size: int | None = None,
        metadata: dict[str, list[str]] | None = None,
    ) -> NativeFile: ...
    def open_append_stream(
        self,
        path: str,
        compression: str | None = "detect",
        buffer_size: int | None = None,
        metadata: dict[str, list[str]] | None = None,
    ): ...
    def normalize_path(self, path: str) -> str: ...

class LocalFileSystem(FileSystem):
    def __init__(self, *, use_mmap: bool = False) -> None: ...

class SubTreeFileSystem(FileSystem):
    def __init__(self, base_path: str, base_fs: FileSystem): ...
    @property
    def base_path(self) -> str: ...
    @property
    def base_fs(self) -> FileSystem: ...

class _MockFileSystem(FileSystem):
    def __init__(self, current_time: dt.datetime | None = None) -> None: ...

class PyFileSystem(FileSystem):
    def __init__(self, handler: FileSystemHandler) -> None: ...
    @property
    def handler(self) -> FileSystemHandler: ...

class FileSystemHandler(ABC):
    @abstractmethod
    def get_type_name(self) -> str: ...
    @abstractmethod
    def get_file_info(self, paths: str | list[str]) -> FileInfo | list[FileInfo]: ...
    @abstractmethod
    def get_file_info_selector(self, selector: FileSelector) -> list[FileInfo]: ...
    @abstractmethod
    def create_dir(self, path: str, recursive: bool) -> None: ...
    @abstractmethod
    def delete_dir(self, path: str) -> None: ...
    @abstractmethod
    def delete_dir_contents(self, path: str, missing_dir_ok: bool = False) -> None: ...
    @abstractmethod
    def delete_root_dir_contents(self) -> None: ...
    @abstractmethod
    def delete_file(self, path: str) -> None: ...
    @abstractmethod
    def move(self, src: str, dest: str) -> None: ...
    @abstractmethod
    def copy_file(self, src: str, dest: str) -> None: ...
    @abstractmethod
    def open_input_stream(self, path: str) -> NativeFile: ...
    @abstractmethod
    def open_input_file(self, path: str) -> NativeFile: ...
    @abstractmethod
    def open_output_stream(self, path: str, metadata: dict[str, list[str]]) -> NativeFile: ...
    @abstractmethod
    def open_append_stream(self, path: str, metadata: dict[str, list[str]]) -> NativeFile: ...
    @abstractmethod
    def normalize_path(self, path: str) -> str: ...



import sys

from pathlib import Path

if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self
from typing import IO, Callable, Iterator, Literal, Sequence

if sys.version_info >= (3, 10):
    from typing import TypeAlias
else:
    from typing_extensions import TypeAlias

from _typeshed import StrPath
from pyarrow import _parquet
from pyarrow._compute import Expression
from pyarrow._fs import FileSystem, SupportedFileSystem
from pyarrow._parquet import (
    ColumnChunkMetaData,
    ColumnSchema,
    FileDecryptionProperties,
    FileEncryptionProperties,
    FileMetaData,
    ParquetLogicalType,
    ParquetReader,
    ParquetSchema,
    RowGroupMetaData,
    SortingColumn,
    Statistics,
)
from pyarrow._stubs_typing import FilterTuple, SingleOrList
from pyarrow.dataset import ParquetFileFragment, Partitioning
from pyarrow.lib import NativeFile, RecordBatch, Schema, Table
from typing_extensions import deprecated

__all__ = (
    "ColumnChunkMetaData",
    "ColumnSchema",
    "FileDecryptionProperties",
    "FileEncryptionProperties",
    "FileMetaData",
    "ParquetDataset",
    "ParquetFile",
    "ParquetLogicalType",
    "ParquetReader",
    "ParquetSchema",
    "ParquetWriter",
    "RowGroupMetaData",
    "SortingColumn",
    "Statistics",
    "read_metadata",
    "read_pandas",
    "read_schema",
    "read_table",
    "write_metadata",
    "write_table",
    "write_to_dataset",
    "_filters_to_expression",
    "filters_to_expression",
)

def filters_to_expression(filters: list[FilterTuple | list[FilterTuple]]) -> Expression: ...
@deprecated("use filters_to_expression")
def _filters_to_expression(filters: list[FilterTuple | list[FilterTuple]]) -> Expression: ...

_Compression: TypeAlias = Literal["gzip", "bz2", "brotli", "lz4", "zstd", "snappy", "none"]

class ParquetFile:
    reader: ParquetReader
    common_metadata: FileMetaData

    def __init__(
        self,
        source: str | Path | NativeFile | IO,
        *,
        metadata: FileMetaData | None = None,
        common_metadata: FileMetaData | None = None,
        read_dictionary: list[str] | None = None,
        memory_map: bool = False,
        buffer_size: int = 0,
        pre_buffer: bool = False,
        coerce_int96_timestamp_unit: str | None = None,
        decryption_properties: FileDecryptionProperties | None = None,
        thrift_string_size_limit: int | None = None,
        thrift_container_size_limit: int | None = None,
        filesystem: SupportedFileSystem | None = None,
        page_checksum_verification: bool = False,
    ): ...
    def __enter__(self) -> Self: ...
    def __exit__(self, *args, **kwargs) -> None: ...
    @property
    def metadata(self) -> FileMetaData: ...
    @property
    def schema(self) -> ParquetSchema: ...
    @property
    def schema_arrow(self) -> Schema: ...
    @property
    def num_row_groups(self) -> int: ...
    def close(self, force: bool = False) -> None: ...
    @property
    def closed(self) -> bool: ...
    def read_row_group(
        self,
        i: int,
        columns: list | None = None,
        use_threads: bool = True,
        use_pandas_metadata: bool = False,
    ) -> Table: ...
    def read_row_groups(
        self,
        row_groups: list,
        columns: list | None = None,
        use_threads: bool = True,
        use_pandas_metadata: bool = False,
    ) -> Table: ...
    def iter_batches(
        self,
        batch_size: int = 65536,
        row_groups: list | None = None,
        columns: list | None = None,
        use_threads: bool = True,
        use_pandas_metadata: bool = False,
    ) -> Iterator[RecordBatch]: ...
    def read(
        self,
        columns: list | None = None,
        use_threads: bool = True,
        use_pandas_metadata: bool = False,
    ) -> Table: ...
    def scan_contents(self, columns: list | None = None, batch_size: int = 65536) -> int: ...

class ParquetWriter:
    flavor: str
    schema_changed: bool
    schema: ParquetSchema
    where: str | Path | IO
    file_handler: NativeFile | None
    writer: _parquet.ParquetWriter
    is_open: bool

    def __init__(
        self,
        where: str | Path | IO | NativeFile,
        schema: Schema,
        filesystem: SupportedFileSystem | None = None,
        flavor: str | None = None,
        version: Literal["1.0", "2.4", "2.6"] = ...,
        use_dictionary: bool = True,
        compression: _Compression | dict[str, _Compression] = "snappy",
        write_statistics: bool | list = True,
        use_deprecated_int96_timestamps: bool | None = None,
        compression_level: int | dict | None = None,
        use_byte_stream_split: bool | list = False,
        column_encoding: str | dict | None = None,
        writer_engine_version=None,
        data_page_version: Literal["1.0", "2.0"] = ...,
        use_compliant_nested_type: bool = True,
        encryption_properties: FileEncryptionProperties | None = None,
        write_batch_size: int | None = None,
        dictionary_pagesize_limit: int | None = None,
        store_schema: bool = True,
        write_page_index: bool = False,
        write_page_checksum: bool = False,
        sorting_columns: Sequence[SortingColumn] | None = None,
        store_decimal_as_integer: bool = False,
        **options,
    ) -> None: ...
    def __enter__(self) -> Self: ...
    def __exit__(self, *args, **kwargs) -> Literal[False]: ...
    def write(
        self, table_or_batch: RecordBatch | Table, row_group_size: int | None = None
    ) -> None: ...
    def write_batch(self, batch: RecordBatch, row_group_size: int | None = None) -> None: ...
    def write_table(self, table: Table, row_group_size: int | None = None) -> None: ...
    def close(self) -> None: ...
    def add_key_value_metadata(self, key_value_metadata: dict[str, str]) -> None: ...

class ParquetDataset:
    def __init__(
        self,
        path_or_paths: SingleOrList[StrPath | NativeFile | IO],
        filesystem: SupportedFileSystem | None = None,
        schema: Schema | None = None,
        *,
        filters: Expression | FilterTuple | list[FilterTuple] | None = None,
        read_dictionary: list[str] | None = None,
        memory_map: bool = False,
        buffer_size: int = 0,
        partitioning: str | list[str] | Partitioning = "hive",
        ignore_prefixes: list[str] | None = None,
        pre_buffer: bool = True,
        coerce_int96_timestamp_unit: str | None = None,
        decryption_properties: FileDecryptionProperties | None = None,
        thrift_string_size_limit: int | None = None,
        thrift_container_size_limit: int | None = None,
        page_checksum_verification: bool = False,
        use_legacy_dataset: bool | None = None,
    ): ...
    def equals(self, other: ParquetDataset) -> bool: ...
    @property
    def schema(self) -> Schema: ...
    def read(
        self,
        columns: list[str] | None = None,
        use_threads: bool = True,
        use_pandas_metadata: bool = False,
    ) -> Table: ...
    def read_pandas(self, **kwargs) -> Table: ...
    @property
    def fragments(self) -> list[ParquetFileFragment]: ...
    @property
    def files(self) -> list[str]: ...
    @property
    def filesystem(self) -> FileSystem: ...
    @property
    def partitioning(self) -> Partitioning: ...

def read_table(
    source: SingleOrList[StrPath | NativeFile | IO],
    *,
    columns: list | None = None,
    use_threads: bool = True,
    schema: Schema | None = None,
    use_pandas_metadata: bool = False,
    read_dictionary: list[str] | None = None,
    memory_map: bool = False,
    buffer_size: int = 0,
    partitioning: str | list[str] | Partitioning = "hive",
    filesystem: SupportedFileSystem | None = None,
    filters: Expression | FilterTuple | list[FilterTuple] | None = None,
    use_legacy_dataset: bool | None = None,
    ignore_prefixes: list[str] | None = None,
    pre_buffer: bool = True,
    coerce_int96_timestamp_unit: str | None = None,
    decryption_properties: FileDecryptionProperties | None = None,
    thrift_string_size_limit: int | None = None,
    thrift_container_size_limit: int | None = None,
    page_checksum_verification: bool = False,
) -> Table: ...
def read_pandas(
    source: str | Path | NativeFile | IO, columns: list | None = None, **kwargs
) -> Table: ...
def write_table(
    table: Table,
    where: str | Path | NativeFile | IO,
    row_group_size: int | None = None,
    version: Literal["1.0", "2.4", "2.6"] = "2.6",
    use_dictionary: bool = True,
    compression: _Compression | dict[str, _Compression] = "snappy",
    write_statistics: bool | list = True,
    use_deprecated_int96_timestamps: bool | None = None,
    coerce_timestamps: str | None = None,
    allow_truncated_timestamps: bool = False,
    data_page_size: int | None = None,
    flavor: str | None = None,
    filesystem: SupportedFileSystem | None = None,
    compression_level: int | dict | None = None,
    use_byte_stream_split: bool = False,
    column_encoding: str | dict | None = None,
    data_page_version: Literal["1.0", "2.0"] = ...,
    use_compliant_nested_type: bool = True,
    encryption_properties: FileEncryptionProperties | None = None,
    write_batch_size: int | None = None,
    dictionary_pagesize_limit: int | None = None,
    store_schema: bool = True,
    write_page_index: bool = False,
    write_page_checksum: bool = False,
    sorting_columns: Sequence[SortingColumn] | None = None,
    store_decimal_as_integer: bool = False,
    **kwargs,
) -> None: ...
def write_to_dataset(
    table: Table,
    root_path: str | Path,
    partition_cols: list[str] | None = None,
    filesystem: SupportedFileSystem | None = None,
    use_legacy_dataset: bool | None = None,
    schema: Schema | None = None,
    partitioning: Partitioning | list[str] | None = None,
    basename_template: str | None = None,
    use_threads: bool | None = None,
    file_visitor: Callable[[str], None] | None = None,
    existing_data_behavior: Literal["overwrite_or_ignore", "error", "delete_matching"]
    | None = None,
    **kwargs,
) -> None: ...
def write_metadata(
    schema: Schema,
    where: str | NativeFile,
    metadata_collector: list[FileMetaData] | None = None,
    filesystem: SupportedFileSystem | None = None,
    **kwargs,
) -> None: ...
def read_metadata(
    where: str | Path | IO | NativeFile,
    memory_map: bool = False,
    decryption_properties: FileDecryptionProperties | None = None,
    filesystem: SupportedFileSystem | None = None,
) -> FileMetaData: ...
def read_schema(
    where: str | Path | IO | NativeFile,
    memory_map: bool = False,
    decryption_properties: FileDecryptionProperties | None = None,
    filesystem: SupportedFileSystem | None = None,
) -> Schema: ...



import datetime as dt
import sys

from collections.abc import Mapping
from decimal import Decimal

if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self

from typing import Any, Generic, Iterable, Iterator, Literal, overload

import numpy as np
import pandas as pd

from pyarrow._stubs_typing import SupportArrowSchema
from pyarrow.lib import (
    Array,
    ChunkedArray,
    ExtensionArray,
    MemoryPool,
    MonthDayNano,
    Table,
)
from typing_extensions import TypeVar

from .io import Buffer
from .scalar import ExtensionScalar

_AsPyType = TypeVar("_AsPyType")
_DataTypeT = TypeVar("_DataTypeT", bound=DataType)
_DataType_CoT = TypeVar("_DataType_CoT", bound=DataType, covariant=True)

class _Weakrefable: ...
class _Metadata(_Weakrefable): ...

class DataType(_Weakrefable):
    def field(self, i: int) -> Field: ...
    @property
    def id(self) -> int: ...
    @property
    def bit_width(self) -> int: ...
    @property
    def byte_width(self) -> int: ...
    @property
    def num_fields(self) -> int: ...
    @property
    def num_buffers(self) -> int: ...
    def __hash__(self) -> int: ...
    def equals(self, other: DataType | str, *, check_metadata: bool = False) -> bool: ...
    def to_pandas_dtype(self) -> np.generic: ...
    def _export_to_c(self, out_ptr: int) -> None: ...
    @classmethod
    def _import_from_c(cls, in_ptr: int) -> Self: ...
    def __arrow_c_schema__(self) -> Any: ...
    @classmethod
    def _import_from_c_capsule(cls, schema) -> Self: ...

class _BasicDataType(DataType, Generic[_AsPyType]): ...
class NullType(_BasicDataType[None]): ...
class BoolType(_BasicDataType[bool]): ...
class Uint8Type(_BasicDataType[int]): ...
class Int8Type(_BasicDataType[int]): ...
class Uint16Type(_BasicDataType[int]): ...
class Int16Type(_BasicDataType[int]): ...
class Uint32Type(_BasicDataType[int]): ...
class Int32Type(_BasicDataType[int]): ...
class Uint64Type(_BasicDataType[int]): ...
class Int64Type(_BasicDataType[int]): ...
class Float16Type(_BasicDataType[float]): ...
class Float32Type(_BasicDataType[float]): ...
class Float64Type(_BasicDataType[float]): ...
class Date32Type(_BasicDataType[dt.date]): ...
class Date64Type(_BasicDataType[dt.date]): ...
class MonthDayNanoIntervalType(_BasicDataType[MonthDayNano]): ...
class StringType(_BasicDataType[str]): ...
class LargeStringType(_BasicDataType[str]): ...
class StringViewType(_BasicDataType[str]): ...
class BinaryType(_BasicDataType[bytes]): ...
class LargeBinaryType(_BasicDataType[bytes]): ...
class BinaryViewType(_BasicDataType[bytes]): ...

_Unit = TypeVar("_Unit", bound=Literal["s", "ms", "us", "ns"])
_Tz = TypeVar("_Tz", str, None, default=None)

class TimestampType(_BasicDataType[int], Generic[_Unit, _Tz]):
    @property
    def unit(self) -> _Unit: ...
    @property
    def tz(self) -> _Tz: ...

_Time32Unit = TypeVar("_Time32Unit", bound=Literal["s", "ms"])

class Time32Type(_BasicDataType[dt.time], Generic[_Time32Unit]):
    @property
    def unit(self) -> _Time32Unit: ...

_Time64Unit = TypeVar("_Time64Unit", bound=Literal["us", "ns"])

class Time64Type(_BasicDataType[dt.time], Generic[_Time64Unit]):
    @property
    def unit(self) -> _Time64Unit: ...

class DurationType(_BasicDataType[dt.timedelta], Generic[_Unit]):
    @property
    def unit(self) -> _Unit: ...

class FixedSizeBinaryType(_BasicDataType[Decimal]): ...

_Precision = TypeVar("_Precision")
_Scale = TypeVar("_Scale")

class Decimal128Type(FixedSizeBinaryType, Generic[_Precision, _Scale]):
    @property
    def precision(self) -> _Precision: ...
    @property
    def scale(self) -> _Scale: ...

class Decimal256Type(FixedSizeBinaryType, Generic[_Precision, _Scale]):
    @property
    def precision(self) -> _Precision: ...
    @property
    def scale(self) -> _Scale: ...

class ListType(DataType, Generic[_DataType_CoT]):
    @property
    def value_field(self) -> Field[_DataType_CoT]: ...
    @property
    def value_type(self) -> _DataType_CoT: ...

class LargeListType(ListType[_DataType_CoT]): ...
class ListViewType(ListType[_DataType_CoT]): ...
class LargeListViewType(ListType[_DataType_CoT]): ...

class FixedSizeListType(ListType[_DataType_CoT], Generic[_DataType_CoT, _Size]):
    @property
    def list_size(self) -> _Size: ...

class DictionaryMemo(_Weakrefable): ...

_IndexT = TypeVar(
    "_IndexT",
    Uint8Type,
    Int8Type,
    Uint16Type,
    Int16Type,
    Uint32Type,
    Int32Type,
    Uint64Type,
    Int64Type,
)
_BasicValueT = TypeVar("_BasicValueT", bound=_BasicDataType)
_ValueT = TypeVar("_ValueT", bound=DataType)
_Ordered = TypeVar("_Ordered", bound=Literal[True, False], default=Literal[False])

class DictionaryType(DataType, Generic[_IndexT, _BasicValueT, _Ordered]):
    @property
    def ordered(self) -> _Ordered: ...
    @property
    def index_type(self) -> _IndexT: ...
    @property
    def value_type(self) -> _BasicValueT: ...

_K = TypeVar("_K", bound=DataType)

class MapType(DataType, Generic[_K, _ValueT, _Ordered]):
    @property
    def key_field(self) -> Field[_K]: ...
    @property
    def key_type(self) -> _K: ...
    @property
    def item_field(self) -> Field[_ValueT]: ...
    @property
    def item_type(self) -> _ValueT: ...
    @property
    def keys_sorted(self) -> _Ordered: ...

_Size = TypeVar("_Size")

class StructType(DataType):
    def get_field_index(self, name: str) -> int: ...
    def field(self, i: int | str) -> Field: ...
    def get_all_field_indices(self, name: str) -> list[int]: ...
    def __len__(self) -> int: ...
    def __iter__(self) -> Iterator[Field]: ...
    __getitem__ = field

class UnionType(DataType):
    @property
    def mode(self) -> Literal["sparse", "dense"]: ...
    @property
    def type_codes(self) -> list[int]: ...
    def __len__(self) -> int: ...
    def __iter__(self) -> Iterator[Field]: ...
    def field(self, i: int) -> Field: ...
    __getitem__ = field

class SparseUnionType(UnionType):
    @property
    def mode(self) -> Literal["sparse"]: ...

class DenseUnionType(UnionType):
    @property
    def mode(self) -> Literal["dense"]: ...

_RunEndType = TypeVar("_RunEndType", Int16Type, Int32Type, Int64Type)

class RunEndEncodedType(DataType, Generic[_RunEndType, _BasicValueT]):
    @property
    def run_end_type(self) -> _RunEndType: ...
    @property
    def value_type(self) -> _BasicValueT: ...

_StorageT = TypeVar("_StorageT", bound=Array | ChunkedArray)

class BaseExtensionType(DataType):
    def __arrow_ext_class__(self) -> type[ExtensionArray]: ...
    def __arrow_ext_scalar_class__(self) -> type[ExtensionScalar]: ...
    @property
    def extension_name(self) -> str: ...
    @property
    def storage_type(self) -> DataType: ...
    def wrap_array(self, storage: _StorageT) -> _StorageT: ...

class ExtensionType(BaseExtensionType):
    def __init__(self, storage_type: DataType, extension_name: str) -> None: ...
    def __arrow_ext_serialize__(self) -> bytes: ...
    @classmethod
    def __arrow_ext_deserialize__(cls, storage_type: DataType, serialized: bytes) -> Self: ...

class FixedShapeTensorType(BaseExtensionType, Generic[_ValueT]):
    @property
    def value_type(self) -> _ValueT: ...
    @property
    def shape(self) -> list[int]: ...
    @property
    def dim_names(self) -> list[str] | None: ...
    @property
    def permutation(self) -> list[int] | None: ...

class PyExtensionType(ExtensionType):
    def __init__(self, storage_type: DataType) -> None: ...
    @classmethod
    def set_auto_load(cls, value: bool) -> None: ...

class UnknownExtensionType(PyExtensionType):
    def __init__(self, storage_type: DataType, serialized: bytes) -> None: ...

def register_extension_type(ext_type: PyExtensionType) -> None: ...
def unregister_extension_type(type_name: str) -> None: ...

class KeyValueMetadata(_Metadata, Mapping[bytes, bytes]):
    def __init__(self, __arg0__: Mapping[bytes, bytes] | None = None, **kwargs) -> None: ...
    def equals(self, other: KeyValueMetadata) -> bool: ...
    def __len__(self) -> int: ...
    def __contains__(self, __key: object) -> bool: ...
    def __getitem__(self, __key: Any) -> Any: ...
    def __iter__(self) -> Iterator[bytes]: ...
    def get_all(self, key: str) -> list[bytes]: ...
    def to_dict(self) -> dict[bytes, bytes]: ...

def ensure_metadata(
    meta: Mapping[bytes | str, bytes | str] | KeyValueMetadata | None, allow_none: bool = False
) -> KeyValueMetadata | None: ...

class Field(_Weakrefable, Generic[_DataType_CoT]):
    def equals(self, other: Field, check_metadata: bool = False) -> bool: ...
    def __hash__(self) -> int: ...
    @property
    def nullable(self) -> bool: ...
    @property
    def name(self) -> str: ...
    @property
    def metadata(self) -> dict[bytes, bytes] | None: ...
    @property
    def type(self) -> _DataType_CoT: ...
    def with_metadata(self, metadata: dict[bytes | str, bytes | str]) -> Self: ...
    def remove_metadata(self) -> None: ...
    def with_type(self, new_type: _DataTypeT) -> Field[_DataTypeT]: ...
    def with_name(self, name: str) -> Self: ...
    def with_nullable(self, nullable: bool) -> Field[_DataType_CoT]: ...
    def flatten(self) -> list[Field]: ...
    def _export_to_c(self, out_ptr: int) -> None: ...
    @classmethod
    def _import_from_c(cls, in_ptr: int) -> Self: ...
    def __arrow_c_schema__(self) -> Any: ...
    @classmethod
    def _import_from_c_capsule(cls, schema) -> Self: ...

class Schema(_Weakrefable):
    def __len__(self) -> int: ...
    def __getitem__(self, key: str) -> Field: ...
    _field = __getitem__
    def __iter__(self) -> Iterator[Field]: ...
    def __hash__(self) -> int: ...
    def __sizeof__(self) -> int: ...
    @property
    def pandas_metadata(self) -> dict: ...
    @property
    def names(self) -> list[str]: ...
    @property
    def types(self) -> list[DataType]: ...
    @property
    def metadata(self) -> dict[bytes, bytes]: ...
    def empty_table(self) -> Table: ...
    def equals(self, other: Schema, check_metadata: bool = False) -> bool: ...
    @classmethod
    def from_pandas(cls, df: pd.DataFrame, preserve_index: bool | None = None) -> Schema: ...
    def field(self, i: int | str | bytes) -> Field: ...
    def field_by_name(self, name: str) -> Field: ...
    def get_field_index(self, name: str) -> int: ...
    def get_all_field_indices(self, name: str) -> list[int]: ...
    def append(self, field: Field) -> Schema: ...
    def insert(self, i: int, field: Field) -> Schema: ...
    def remove(self, i: int) -> Schema: ...
    def set(self, i: int, field: Field) -> Schema: ...
    def add_metadata(self, metadata: dict) -> Schema: ...
    def with_metadata(self, metadata: dict) -> Schema: ...
    def serialize(self, memory_pool: MemoryPool | None = None) -> Buffer: ...
    def remove_metadata(self) -> Schema: ...
    def to_string(
        self,
        truncate_metadata: bool = True,
        show_field_metadata: bool = True,
        show_schema_metadata: bool = True,
    ) -> str: ...
    def _export_to_c(self, out_ptr: int) -> None: ...
    @classmethod
    def _import_from_c(cls, in_ptr: int) -> Schema: ...
    def __arrow_c_schema__(self) -> Any: ...
    @staticmethod
    def _import_from_c_capsule(schema: Any) -> Schema: ...

def unify_schemas(
    schemas: list[Schema], *, promote_options: Literal["default", "permissive"] = "default"
) -> Schema: ...
@overload
def field(name: SupportArrowSchema) -> Field: ...
@overload
def field(
    name: str, type: _DataTypeT, nullable: bool = ..., metadata: dict | None = None
) -> Field[_DataTypeT]: ...
def null() -> NullType: ...
def bool_() -> BoolType: ...
def uint8() -> Uint8Type: ...
def int8() -> Int8Type: ...
def uint16() -> Uint16Type: ...
def int16() -> Int16Type: ...
def uint32() -> Uint32Type: ...
def int32() -> Int32Type: ...
def int64() -> Int64Type: ...
def uint64() -> Uint64Type: ...
def tzinfo_to_string(tz: dt.tzinfo) -> str: ...
def string_to_tzinfo(name: str) -> dt.tzinfo: ...
@overload
def timestamp(unit: _Unit) -> TimestampType[_Unit, None]: ...
@overload
def timestamp(unit: _Unit, tz: _Tz) -> TimestampType[_Unit, _Tz]: ...
def time32(unit: _Time32Unit) -> Time32Type[_Time32Unit]: ...
def time64(unit: _Time64Unit) -> Time64Type[_Time64Unit]: ...
def duration(unit: _Unit) -> DurationType[_Unit]: ...
def month_day_nano_interval() -> MonthDayNanoIntervalType: ...
def date32() -> Date32Type: ...
def date64() -> Date64Type: ...
def float16() -> Float16Type: ...
def float32() -> Float32Type: ...
def float64() -> Float64Type: ...
@overload
def decimal128(precision: _Precision) -> Decimal128Type[_Precision, Literal[0]]: ...
@overload
def decimal128(precision: _Precision, scale: _Scale) -> Decimal128Type[_Precision, _Scale]: ...
@overload
def decimal256(precision: _Precision) -> Decimal256Type[_Precision, Literal[0]]: ...
@overload
def decimal256(precision: _Precision, scale: _Scale) -> Decimal256Type[_Precision, _Scale]: ...
def string() -> StringType: ...

utf8 = string

@overload
def binary() -> BinaryType: ...
@overload
def binary(length: Literal[-1]) -> BinaryType: ...  # type: ignore[overload-overlap]
@overload
def binary(length: int) -> FixedSizeBinaryType: ...
def large_binary() -> LargeBinaryType: ...
def large_string() -> LargeStringType: ...

large_utf8 = large_string

def binary_view() -> BinaryViewType: ...
def string_view() -> StringViewType: ...
@overload
def list_(value_type: Field[_DataTypeT]) -> ListType[_DataTypeT]: ...
@overload
def list_(value_type: _DataTypeT) -> ListType[_DataTypeT]: ...
@overload
def list_(value_type: _DataTypeT, list_size: Literal[-1]) -> ListType[_DataTypeT]: ...  # type: ignore[overload-overlap]
@overload
def list_(value_type: _DataTypeT, list_size: _Size) -> FixedSizeListType[_DataTypeT, _Size]: ...
@overload
def large_list(value_type: Field[_DataTypeT]) -> LargeListType[_DataTypeT]: ...
@overload
def large_list(value_type: _DataTypeT) -> LargeListType[_DataTypeT]: ...
@overload
def list_view(value_type: Field[_DataTypeT]) -> ListViewType[_DataTypeT]: ...
@overload
def list_view(value_type: _DataTypeT) -> ListViewType[_DataTypeT]: ...
@overload
def large_list_view(value_type: Field[_DataTypeT]) -> LargeListViewType[_DataTypeT]: ...
@overload
def large_list_view(value_type: _DataTypeT) -> LargeListViewType[_DataTypeT]: ...
@overload
def map_(key_type: _K, item_type: _ValueT) -> MapType[_K, _ValueT, Literal[False]]: ...
@overload
def map_(
    key_type: _K, item_type: _ValueT, key_sorted: _Ordered
) -> MapType[_K, _ValueT, _Ordered]: ...
@overload
def dictionary(
    index_type: _IndexT, value_type: _BasicValueT
) -> DictionaryType[_IndexT, _BasicValueT, Literal[False]]: ...
@overload
def dictionary(
    index_type: _IndexT, value_type: _BasicValueT, ordered: _Ordered
) -> DictionaryType[_IndexT, _BasicValueT, _Ordered]: ...
def struct(
    fields: Iterable[Field | tuple[str, Field]] | Mapping[str, Field],
) -> StructType: ...
def sparse_union(
    child_fields: list[Field], type_codes: list[int] | None = None
) -> SparseUnionType: ...
def dense_union(
    child_fields: list[Field], type_codes: list[int] | None = None
) -> DenseUnionType: ...
@overload
def union(
    child_fields: list[Field], mode: Literal["sparse"], type_codes: list[int] | None = None
) -> SparseUnionType: ...
@overload
def union(
    child_fields: list[Field], mode: Literal["dense"], type_codes: list[int] | None = None
) -> DenseUnionType: ...
def run_end_encoded(
    run_end_type: _RunEndType, value_type: _BasicValueT
) -> RunEndEncodedType[_RunEndType, _BasicValueT]: ...
def fixed_shape_tensor(
    value_type: _ValueT,
    shape: tuple[list[int], ...],
    dim_names: tuple[list[str], ...] | None = None,
    permutation: tuple[list[int], ...] | None = None,
) -> FixedShapeTensorType[_ValueT]: ...
@overload
def type_for_alias(name: Literal["null"]) -> NullType: ...
@overload
def type_for_alias(name: Literal["bool", "boolean"]) -> BoolType: ...
@overload
def type_for_alias(name: Literal["i1", "int8"]) -> Int8Type: ...
@overload
def type_for_alias(name: Literal["i2", "int16"]) -> Int16Type: ...
@overload
def type_for_alias(name: Literal["i4", "int32"]) -> Int32Type: ...
@overload
def type_for_alias(name: Literal["i8", "int64"]) -> Int64Type: ...
@overload
def type_for_alias(name: Literal["u1", "uint8"]) -> Uint8Type: ...
@overload
def type_for_alias(name: Literal["u2", "uint16"]) -> Uint16Type: ...
@overload
def type_for_alias(name: Literal["u4", "uint32"]) -> Uint32Type: ...
@overload
def type_for_alias(name: Literal["u8", "uint64"]) -> Uint64Type: ...
@overload
def type_for_alias(name: Literal["f2", "halffloat", "float16"]) -> Float16Type: ...
@overload
def type_for_alias(name: Literal["f4", "float", "float32"]) -> Float32Type: ...
@overload
def type_for_alias(name: Literal["f8", "double", "float64"]) -> Float64Type: ...
@overload
def type_for_alias(name: Literal["string", "str", "utf8"]) -> StringType: ...
@overload
def type_for_alias(name: Literal["binary"]) -> BinaryType: ...
@overload
def type_for_alias(
    name: Literal["large_string", "large_str", "large_utf8"],
) -> LargeStringType: ...
@overload
def type_for_alias(name: Literal["large_binary"]) -> LargeBinaryType: ...
@overload
def type_for_alias(name: Literal["binary_view"]) -> BinaryViewType: ...
@overload
def type_for_alias(name: Literal["string_view"]) -> StringViewType: ...
@overload
def type_for_alias(name: Literal["date32", "date32[day]"]) -> Date32Type: ...
@overload
def type_for_alias(name: Literal["date64", "date64[ms]"]) -> Date64Type: ...
@overload
def type_for_alias(name: Literal["time32[s]"]) -> Time32Type[Literal["s"]]: ...
@overload
def type_for_alias(name: Literal["time32[ms]"]) -> Time32Type[Literal["ms"]]: ...
@overload
def type_for_alias(name: Literal["time64[us]"]) -> Time64Type[Literal["us"]]: ...
@overload
def type_for_alias(name: Literal["time64[ns]"]) -> Time64Type[Literal["ns"]]: ...
@overload
def type_for_alias(name: Literal["timestamp[s]"]) -> TimestampType[Literal["s"], Any]: ...
@overload
def type_for_alias(name: Literal["timestamp[ms]"]) -> TimestampType[Literal["ms"], Any]: ...
@overload
def type_for_alias(name: Literal["timestamp[us]"]) -> TimestampType[Literal["us"], Any]: ...
@overload
def type_for_alias(name: Literal["timestamp[ns]"]) -> TimestampType[Literal["ns"], Any]: ...
@overload
def type_for_alias(name: Literal["duration[s]"]) -> DurationType[Literal["s"]]: ...
@overload
def type_for_alias(name: Literal["duration[ms]"]) -> DurationType[Literal["ms"]]: ...
@overload
def type_for_alias(name: Literal["duration[us]"]) -> DurationType[Literal["us"]]: ...
@overload
def type_for_alias(name: Literal["duration[ns]"]) -> DurationType[Literal["ns"]]: ...
@overload
def type_for_alias(name: Literal["month_day_nano_interval"]) -> MonthDayNanoIntervalType: ...
@overload
def ensure_type(ty: None, allow_none: Literal[True]) -> None: ...
@overload
def ensure_type(ty: _DataTypeT) -> _DataTypeT: ...
@overload
def ensure_type(ty: Literal["null"]) -> NullType: ...
@overload
def ensure_type(ty: Literal["bool", "boolean"]) -> BoolType: ...
@overload
def ensure_type(ty: Literal["i1", "int8"]) -> Int8Type: ...
@overload
def ensure_type(ty: Literal["i2", "int16"]) -> Int16Type: ...
@overload
def ensure_type(ty: Literal["i4", "int32"]) -> Int32Type: ...
@overload
def ensure_type(ty: Literal["i8", "int64"]) -> Int64Type: ...
@overload
def ensure_type(ty: Literal["u1", "uint8"]) -> Uint8Type: ...
@overload
def ensure_type(ty: Literal["u2", "uint16"]) -> Uint16Type: ...
@overload
def ensure_type(ty: Literal["u4", "uint32"]) -> Uint32Type: ...
@overload
def ensure_type(ty: Literal["u8", "uint64"]) -> Uint64Type: ...
@overload
def ensure_type(ty: Literal["f2", "halffloat", "float16"]) -> Float16Type: ...
@overload
def ensure_type(ty: Literal["f4", "float", "float32"]) -> Float32Type: ...
@overload
def ensure_type(ty: Literal["f8", "double", "float64"]) -> Float64Type: ...
@overload
def ensure_type(ty: Literal["string", "str", "utf8"]) -> StringType: ...
@overload
def ensure_type(ty: Literal["binary"]) -> BinaryType: ...
@overload
def ensure_type(
    ty: Literal["large_string", "large_str", "large_utf8"],
) -> LargeStringType: ...
@overload
def ensure_type(ty: Literal["large_binary"]) -> LargeBinaryType: ...
@overload
def ensure_type(ty: Literal["binary_view"]) -> BinaryViewType: ...
@overload
def ensure_type(ty: Literal["string_view"]) -> StringViewType: ...
@overload
def ensure_type(ty: Literal["date32", "date32[day]"]) -> Date32Type: ...
@overload
def ensure_type(ty: Literal["date64", "date64[ms]"]) -> Date64Type: ...
@overload
def ensure_type(ty: Literal["time32[s]"]) -> Time32Type[Literal["s"]]: ...
@overload
def ensure_type(ty: Literal["time32[ms]"]) -> Time32Type[Literal["ms"]]: ...
@overload
def ensure_type(ty: Literal["time64[us]"]) -> Time64Type[Literal["us"]]: ...
@overload
def ensure_type(ty: Literal["time64[ns]"]) -> Time64Type[Literal["ns"]]: ...
@overload
def ensure_type(ty: Literal["timestamp[s]"]) -> TimestampType[Literal["s"], Any]: ...
@overload
def ensure_type(ty: Literal["timestamp[ms]"]) -> TimestampType[Literal["ms"], Any]: ...
@overload
def ensure_type(ty: Literal["timestamp[us]"]) -> TimestampType[Literal["us"], Any]: ...
@overload
def ensure_type(ty: Literal["timestamp[ns]"]) -> TimestampType[Literal["ns"], Any]: ...
@overload
def ensure_type(ty: Literal["duration[s]"]) -> DurationType[Literal["s"]]: ...
@overload
def ensure_type(ty: Literal["duration[ms]"]) -> DurationType[Literal["ms"]]: ...
@overload
def ensure_type(ty: Literal["duration[us]"]) -> DurationType[Literal["us"]]: ...
@overload
def ensure_type(ty: Literal["duration[ns]"]) -> DurationType[Literal["ns"]]: ...
@overload
def ensure_type(ty: Literal["month_day_nano_interval"]) -> MonthDayNanoIntervalType: ...
def schema(
    fields: Iterable[Field] | Iterable[tuple[str, DataType]] | Mapping[str, DataType],
    metadata: dict[bytes | str, bytes | str] | None = None,
) -> Schema: ...
def from_numpy_dtype(dtype: np.dtype) -> DataType: ...
def is_boolean_value(obj: Any) -> bool: ...
def is_integer_value(obj: Any) -> bool: ...
def is_float_value(obj: Any) -> bool: ...

__all__ = [
    "_Weakrefable",
    "_Metadata",
    "DataType",
    "_BasicDataType",
    "NullType",
    "BoolType",
    "Uint8Type",
    "Int8Type",
    "Uint16Type",
    "Int16Type",
    "Uint32Type",
    "Int32Type",
    "Uint64Type",
    "Int64Type",
    "Float16Type",
    "Float32Type",
    "Float64Type",
    "Date32Type",
    "Date64Type",
    "MonthDayNanoIntervalType",
    "StringType",
    "LargeStringType",
    "StringViewType",
    "BinaryType",
    "LargeBinaryType",
    "BinaryViewType",
    "TimestampType",
    "Time32Type",
    "Time64Type",
    "DurationType",
    "FixedSizeBinaryType",
    "Decimal128Type",
    "Decimal256Type",
    "ListType",
    "LargeListType",
    "ListViewType",
    "LargeListViewType",
    "FixedSizeListType",
    "DictionaryMemo",
    "DictionaryType",
    "MapType",
    "StructType",
    "UnionType",
    "SparseUnionType",
    "DenseUnionType",
    "RunEndEncodedType",
    "BaseExtensionType",
    "ExtensionType",
    "FixedShapeTensorType",
    "PyExtensionType",
    "UnknownExtensionType",
    "register_extension_type",
    "unregister_extension_type",
    "KeyValueMetadata",
    "ensure_metadata",
    "Field",
    "Schema",
    "unify_schemas",
    "field",
    "null",
    "bool_",
    "uint8",
    "int8",
    "uint16",
    "int16",
    "uint32",
    "int32",
    "int64",
    "uint64",
    "tzinfo_to_string",
    "string_to_tzinfo",
    "timestamp",
    "time32",
    "time64",
    "duration",
    "month_day_nano_interval",
    "date32",
    "date64",
    "float16",
    "float32",
    "float64",
    "decimal128",
    "decimal256",
    "string",
    "utf8",
    "binary",
    "large_binary",
    "large_string",
    "large_utf8",
    "binary_view",
    "string_view",
    "list_",
    "large_list",
    "list_view",
    "large_list_view",
    "map_",
    "dictionary",
    "struct",
    "sparse_union",
    "dense_union",
    "union",
    "run_end_encoded",
    "fixed_shape_tensor",
    "type_for_alias",
    "ensure_type",
    "schema",
    "from_numpy_dtype",
    "is_boolean_value",
    "is_integer_value",
    "is_float_value",
]



from ._dataset_parquet import ParquetFileWriteOptions, ParquetFragmentScanOptions
from ._parquet import FileDecryptionProperties
from ._parquet_encryption import CryptoFactory, EncryptionConfiguration, KmsConnectionConfig
from .lib import _Weakrefable

class ParquetEncryptionConfig(_Weakrefable):
    def __init__(
        self,
        crypto_factory: CryptoFactory,
        kms_connection_config: KmsConnectionConfig,
        encryption_config: EncryptionConfiguration,
    ) -> None: ...

class ParquetDecryptionConfig(_Weakrefable):
    def __init__(
        self,
        crypto_factory: CryptoFactory,
        kms_connection_config: KmsConnectionConfig,
        encryption_config: EncryptionConfiguration,
    ) -> None: ...

def set_encryption_config(
    opts: ParquetFileWriteOptions,
    config: ParquetEncryptionConfig,
) -> None: ...
def set_decryption_properties(
    opts: ParquetFragmentScanOptions,
    config: FileDecryptionProperties,
): ...
def set_decryption_config(
    opts: ParquetFragmentScanOptions,
    config: ParquetDecryptionConfig,
): ...



from typing import IO, Literal

import pandas as pd

from _typeshed import StrPath
from pyarrow._feather import FeatherError
from pyarrow.lib import Table

__all__ = [
    "FeatherError",
    "FeatherDataset",
    "check_chunked_overflow",
    "write_feather",
    "read_feather",
    "read_table",
]

class FeatherDataset:
    path_or_paths: str | list[str]
    validate_schema: bool

    def __init__(self, path_or_paths: str | list[str], validate_schema: bool = True) -> None: ...
    def read_table(self, columns: list[str] | None = None) -> Table: ...
    def validate_schemas(self, piece, table: Table) -> None: ...
    def read_pandas(
        self, columns: list[str] | None = None, use_threads: bool = True
    ) -> pd.DataFrame: ...

def check_chunked_overflow(name: str, col) -> None: ...
def write_feather(
    df: pd.DataFrame | Table,
    dest: StrPath | IO,
    compression: Literal["zstd", "lz4", "uncompressed"] | None = None,
    compression_level: int | None = None,
    chunksize: int | None = None,
    version: Literal[1, 2] = 2,
) -> None: ...
def read_feather(
    source: StrPath | IO,
    columns: list[str] | None = None,
    use_threads: bool = True,
    memory_map: bool = False,
    **kwargs,
) -> pd.DataFrame: ...
def read_table(
    source: StrPath | IO,
    columns: list[str] | None = None,
    memory_map: bool = False,
    use_threads: bool = True,
) -> Table: ...



import datetime as dt

from typing import Callable

from ._parquet import FileDecryptionProperties, FileEncryptionProperties
from .lib import _Weakrefable

class EncryptionConfiguration(_Weakrefable):
    footer_key: str
    column_keys: dict[str, list[str]]
    encryption_algorithm: str
    plaintext_footer: bool
    double_wrapping: bool
    cache_lifetime: dt.timedelta
    internal_key_material: bool
    data_key_length_bits: int

    def __init__(
        self,
        footer_key: str,
        *,
        column_keys: dict[str, str | list[str]] | None = None,
        encryption_algorithm: str | None = None,
        plaintext_footer: bool | None = None,
        double_wrapping: bool | None = None,
        cache_lifetime: dt.timedelta | None = None,
        internal_key_material: bool | None = None,
        data_key_length_bits: int | None = None,
    ) -> None: ...

class DecryptionConfiguration(_Weakrefable):
    cache_lifetime: dt.timedelta
    def __init__(self, *, cache_lifetime: dt.timedelta | None = None): ...

class KmsConnectionConfig(_Weakrefable):
    kms_instance_id: str
    kms_instance_url: str
    key_access_token: str
    custom_kms_conf: dict[str, str]
    def __init__(
        self,
        *,
        kms_instance_id: str | None = None,
        kms_instance_url: str | None = None,
        key_access_token: str | None = None,
        custom_kms_conf: dict[str, str] | None = None,
    ) -> None: ...
    def refresh_key_access_token(self, value: str) -> None: ...

class KmsClient(_Weakrefable):
    def wrap_key(self, key_bytes: bytes, master_key_identifier: str) -> str: ...
    def unwrap_key(self, wrapped_key: str, master_key_identifier: str) -> str: ...

class CryptoFactory(_Weakrefable):
    def __init__(self, kms_client_factory: Callable[[KmsConnectionConfig], KmsClient]): ...
    def file_encryption_properties(
        self,
        kms_connection_config: KmsConnectionConfig,
        encryption_config: EncryptionConfiguration,
    ) -> FileEncryptionProperties: ...
    def file_decryption_properties(
        self,
        kms_connection_config: KmsConnectionConfig,
        decryption_config: DecryptionConfiguration | None = None,
    ) -> FileDecryptionProperties: ...
    def remove_cache_entries_for_token(self, access_token: str) -> None: ...
    def remove_cache_entries_for_all_tokens(self) -> None: ...



from typing import Callable

from ._compute import Expression
from .lib import Buffer, RecordBatchReader, Schema, Table, _Weakrefable

def run_query(
    plan: Buffer | int,
    *,
    table_provider: Callable[[list[str], Schema], Table] | None = None,
    use_threads: bool = True,
) -> RecordBatchReader: ...
def _parse_json_plan(plan: bytes) -> Buffer: ...
def serialize_expressions(
    exprs: list[Expression],
    names: list[str],
    schema: Schema,
    *,
    allow_arrow_extensions: bool = False,
) -> Buffer: ...

class BoundExpressions(_Weakrefable):
    @property
    def schema(self) -> Schema: ...
    @property
    def expressions(self) -> dict[str, Expression]: ...

def deserialize_expressions(buf: Buffer | bytes) -> BoundExpressions: ...
def get_supported_functions() -> list[str]: ...



# ruff: noqa: F403
from typing import NamedTuple

from .__lib_pxi.array import *
from .__lib_pxi.benchmark import *
from .__lib_pxi.builder import *
from .__lib_pxi.compat import *
from .__lib_pxi.config import *
from .__lib_pxi.device import *
from .__lib_pxi.error import *
from .__lib_pxi.io import *
from .__lib_pxi.ipc import *
from .__lib_pxi.memory import *
from .__lib_pxi.pandas_shim import *
from .__lib_pxi.scalar import *
from .__lib_pxi.table import *
from .__lib_pxi.tensor import *
from .__lib_pxi.types import *

class MonthDayNano(NamedTuple):
    days: int
    months: int
    nanoseconds: int

def cpu_count() -> int: ...
def set_cpu_count(count: int) -> None: ...
def is_threading_enabled() -> bool: ...

Type_NA: int
Type_BOOL: int
Type_UINT8: int
Type_INT8: int
Type_UINT16: int
Type_INT16: int
Type_UINT32: int
Type_INT32: int
Type_UINT64: int
Type_INT64: int
Type_HALF_FLOAT: int
Type_FLOAT: int
Type_DOUBLE: int
Type_DECIMAL128: int
Type_DECIMAL256: int
Type_DATE32: int
Type_DATE64: int
Type_TIMESTAMP: int
Type_TIME32: int
Type_TIME64: int
Type_DURATION: int
Type_INTERVAL_MONTH_DAY_NANO: int
Type_BINARY: int
Type_STRING: int
Type_LARGE_BINARY: int
Type_LARGE_STRING: int
Type_FIXED_SIZE_BINARY: int
Type_BINARY_VIEW: int
Type_STRING_VIEW: int
Type_LIST: int
Type_LARGE_LIST: int
Type_LIST_VIEW: int
Type_LARGE_LIST_VIEW: int
Type_MAP: int
Type_FIXED_SIZE_LIST: int
Type_STRUCT: int
Type_SPARSE_UNION: int
Type_DENSE_UNION: int
Type_DICTIONARY: int
Type_RUN_END_ENCODED: int
UnionMode_SPARSE: int
UnionMode_DENSE: int



# mypy: disable-error-code="import-untyped"

import sys

if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self

import numpy as np

from pyarrow.lib import _Weakrefable
from pydata.sparse import COO  # type: ignore[import-not-found]
from scipy.sparse import coo_matrix, csr_matrix

class Tensor(_Weakrefable):
    @classmethod
    def from_numpy(cls, obj: np.ndarray, dim_names: list[str] | None = None) -> Self: ...
    def to_numpy(self) -> np.ndarray: ...
    def equals(self, other: Tensor) -> bool: ...
    def dim_name(self, i: int) -> str: ...
    @property
    def dim_names(self) -> list[str]: ...
    @property
    def is_mutable(self) -> bool: ...
    @property
    def is_contiguous(self) -> bool: ...
    @property
    def ndim(self) -> int: ...
    @property
    def size(self) -> str: ...
    @property
    def shape(self) -> tuple[int, ...]: ...
    @property
    def strides(self) -> tuple[int, ...]: ...

class SparseCOOTensor(_Weakrefable):
    @classmethod
    def from_dense_numpy(cls, obj: np.ndarray, dim_names: list[str] | None = None) -> Self: ...
    @classmethod
    def from_numpy(
        cls,
        data: np.ndarray,
        coords: np.ndarray,
        shape: tuple[int, ...],
        dim_names: list[str] | None = None,
    ) -> Self: ...
    @classmethod
    def from_scipy(cls, obj: csr_matrix, dim_names: list[str] | None = None) -> Self: ...
    @classmethod
    def from_pydata_sparse(cls, obj: COO, dim_names: list[str] | None = None) -> Self: ...
    @classmethod
    def from_tensor(cls, obj: Tensor) -> Self: ...
    def to_numpy(self) -> tuple[np.ndarray, np.ndarray]: ...
    def to_scipy(self) -> coo_matrix: ...
    def to_pydata_sparse(self) -> COO: ...
    def to_tensor(self) -> Tensor: ...
    def equals(self, other: Self) -> bool: ...
    @property
    def is_mutable(self) -> bool: ...
    @property
    def ndim(self) -> int: ...
    @property
    def size(self) -> str: ...
    @property
    def shape(self) -> tuple[int, ...]: ...
    def dim_name(self, i: int) -> str: ...
    @property
    def dim_names(self) -> list[str]: ...
    @property
    def non_zero_length(self) -> int: ...
    @property
    def has_canonical_format(self) -> bool: ...

class SparseCSRMatrix(_Weakrefable):
    @classmethod
    def from_dense_numpy(cls, obj: np.ndarray, dim_names: list[str] | None = None) -> Self: ...
    @classmethod
    def from_numpy(
        cls,
        data: np.ndarray,
        indptr: np.ndarray,
        indices: np.ndarray,
        shape: tuple[int, ...],
        dim_names: list[str] | None = None,
    ) -> Self: ...
    @classmethod
    def from_scipy(cls, obj: csr_matrix, dim_names: list[str] | None = None) -> Self: ...
    @classmethod
    def from_tensor(cls, obj: Tensor) -> Self: ...
    def to_numpy(self) -> tuple[np.ndarray, np.ndarray, np.ndarray]: ...
    def to_scipy(self) -> csr_matrix: ...
    def to_tensor(self) -> Tensor: ...
    def equals(self, other: Self) -> bool: ...
    @property
    def is_mutable(self) -> bool: ...
    @property
    def ndim(self) -> int: ...
    @property
    def size(self) -> str: ...
    @property
    def shape(self) -> tuple[int, ...]: ...
    def dim_name(self, i: int) -> str: ...
    @property
    def dim_names(self) -> list[str]: ...
    @property
    def non_zero_length(self) -> int: ...

class SparseCSCMatrix(_Weakrefable):
    @classmethod
    def from_dense_numpy(cls, obj: np.ndarray, dim_names: list[str] | None = None) -> Self: ...
    @classmethod
    def from_numpy(
        cls,
        data: np.ndarray,
        indptr: np.ndarray,
        indices: np.ndarray,
        shape: tuple[int, ...],
        dim_names: list[str] | None = None,
    ) -> Self: ...
    @classmethod
    def from_scipy(cls, obj: csr_matrix, dim_names: list[str] | None = None) -> Self: ...
    @classmethod
    def from_tensor(cls, obj: Tensor) -> Self: ...
    def to_numpy(self) -> tuple[np.ndarray, np.ndarray, np.ndarray]: ...
    def to_scipy(self) -> csr_matrix: ...
    def to_tensor(self) -> Tensor: ...
    def equals(self, other: Self) -> bool: ...
    @property
    def is_mutable(self) -> bool: ...
    @property
    def ndim(self) -> int: ...
    @property
    def size(self) -> str: ...
    @property
    def shape(self) -> tuple[int, ...]: ...
    def dim_name(self, i: int) -> str: ...
    @property
    def dim_names(self) -> list[str]: ...
    @property
    def non_zero_length(self) -> int: ...

class SparseCSFTensor(_Weakrefable):
    @classmethod
    def from_dense_numpy(cls, obj: np.ndarray, dim_names: list[str] | None = None) -> Self: ...
    @classmethod
    def from_numpy(
        cls,
        data: np.ndarray,
        indptr: np.ndarray,
        indices: np.ndarray,
        shape: tuple[int, ...],
        dim_names: list[str] | None = None,
    ) -> Self: ...
    @classmethod
    def from_scipy(cls, obj: csr_matrix, dim_names: list[str] | None = None) -> Self: ...
    @classmethod
    def from_tensor(cls, obj: Tensor) -> Self: ...
    def to_numpy(self) -> tuple[np.ndarray, np.ndarray, np.ndarray]: ...
    def to_tensor(self) -> Tensor: ...
    def equals(self, other: Self) -> bool: ...
    @property
    def is_mutable(self) -> bool: ...
    @property
    def ndim(self) -> int: ...
    @property
    def size(self) -> str: ...
    @property
    def shape(self) -> tuple[int, ...]: ...
    def dim_name(self, i: int) -> str: ...
    @property
    def dim_names(self) -> list[str]: ...
    @property
    def non_zero_length(self) -> int: ...

__all__ = [
    "Tensor",
    "SparseCOOTensor",
    "SparseCSRMatrix",
    "SparseCSCMatrix",
    "SparseCSFTensor",
]



# mypy: disable-error-code="name-defined"
from types import ModuleType
from typing import Any, Iterable, TypeGuard

import pandas as pd

from numpy import dtype
from pandas.core.dtypes.base import ExtensionDtype

class _PandasAPIShim:
    has_sparse: bool

    def series(self, *args, **kwargs) -> pd.Series: ...
    def data_frame(self, *args, **kwargs) -> pd.DataFrame: ...
    @property
    def have_pandas(self) -> bool: ...
    @property
    def compat(self) -> ModuleType: ...
    @property
    def pd(self) -> ModuleType: ...
    def infer_dtype(self, obj: Iterable) -> str: ...
    def pandas_dtype(self, dtype: str) -> dtype: ...
    @property
    def loose_version(self) -> Any: ...
    @property
    def version(self) -> str: ...
    def is_v1(self) -> bool: ...
    def is_ge_v21(self) -> bool: ...
    def is_ge_v3(self) -> bool: ...
    @property
    def categorical_type(self) -> type[pd.Categorical]: ...
    @property
    def datetimetz_type(self) -> type[pd.DatetimeTZDtype]: ...
    @property
    def extension_dtype(self) -> type[ExtensionDtype]: ...
    def is_array_like(
        self, obj: Any
    ) -> TypeGuard[pd.Series | pd.Index | pd.Categorical | ExtensionDtype]: ...
    def is_categorical(self, obj: Any) -> TypeGuard[pd.Categorical]: ...
    def is_datetimetz(self, obj: Any) -> TypeGuard[pd.DatetimeTZDtype]: ...
    def is_extension_array_dtype(self, obj: Any) -> TypeGuard[ExtensionDtype]: ...
    def is_sparse(self, obj: Any) -> bool: ...
    def is_data_frame(self, obj: Any) -> TypeGuard[pd.DataFrame]: ...
    def is_series(self, obj: Any) -> TypeGuard[pd.Series]: ...
    def is_index(self, obj: Any) -> TypeGuard[pd.Index]: ...
    def get_values(self, obj: Any) -> bool: ...
    def get_rangeindex_attribute(self, level, name): ...

_pandas_api: _PandasAPIShim

__all__ = ["_PandasAPIShim", "_pandas_api"]



import enum

from typing import Literal, NotRequired, Required, TypedDict

from ._fs import FileSystem
from .lib import KeyValueMetadata

class _ProxyOptions(TypedDict):
    schema: Required[Literal["http", "https"]]
    host: Required[str]
    port: Required[int]
    username: NotRequired[str]
    password: NotRequired[str]

class S3LogLevel(enum.IntEnum):
    Off = enum.auto()
    Fatal = enum.auto()
    Error = enum.auto()
    Warn = enum.auto()
    Info = enum.auto()
    Debug = enum.auto()
    Trace = enum.auto()

Off = S3LogLevel.Off
Fatal = S3LogLevel.Fatal
Error = S3LogLevel.Error
Warn = S3LogLevel.Warn
Info = S3LogLevel.Info
Debug = S3LogLevel.Debug
Trace = S3LogLevel.Trace

def initialize_s3(
    log_level: S3LogLevel = S3LogLevel.Fatal, num_event_loop_threads: int = 1
) -> None: ...
def ensure_s3_initialized() -> None: ...
def finalize_s3() -> None: ...
def ensure_s3_finalized() -> None: ...
def resolve_s3_region(bucket: str) -> str: ...

class S3RetryStrategy:
    max_attempts: int
    def __init__(self, max_attempts=3) -> None: ...

class AwsStandardS3RetryStrategy(S3RetryStrategy): ...
class AwsDefaultS3RetryStrategy(S3RetryStrategy): ...

class S3FileSystem(FileSystem):
    def __init__(
        self,
        *,
        access_key: str | None = None,
        secret_key: str | None = None,
        session_token: str | None = None,
        anonymous: bool = False,
        region: str | None = None,
        request_timeout: float | None = None,
        connect_timeout: float | None = None,
        scheme: Literal["http", "https"] = "https",
        endpoint_override: str | None = None,
        background_writes: bool = True,
        default_metadata: dict | KeyValueMetadata | None = None,
        role_arn: str | None = None,
        session_name: str | None = None,
        external_id: str | None = None,
        load_frequency: int = 900,
        proxy_options: _ProxyOptions | str | None = None,
        allow_bucket_creation: bool = False,
        allow_bucket_deletion: bool = False,
        check_directory_existence_before_creation: bool = False,
        retry_strategy: S3RetryStrategy = AwsStandardS3RetryStrategy(max_attempts=3),
        force_virtual_addressing: bool = False,
    ): ...
    @property
    def region(self) -> str: ...



from typing import IO

from _typeshed import StrPath

from .lib import Buffer, NativeFile, Table, _Weakrefable

class FeatherError(Exception): ...

def write_feather(
    table: Table,
    dest: StrPath | IO | NativeFile,
    compression: str | None = None,
    compression_level: int | None = None,
    chunksize: int | None = None,
    version: int = 2,
): ...

class FeatherReader(_Weakrefable):
    def __init__(
        self,
        source: StrPath | IO | NativeFile | Buffer,
        use_memory_map: bool,
        use_threads: bool,
    ) -> None: ...
    @property
    def version(self) -> str: ...
    def read(self) -> Table: ...
    def read_indices(self, indices: list[int]) -> Table: ...
    def read_names(self, names: list[str]) -> Table: ...



import sys

if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self

class ArrowException(Exception): ...
class ArrowInvalid(ValueError, ArrowException): ...
class ArrowMemoryError(MemoryError, ArrowException): ...
class ArrowKeyError(KeyError, ArrowException): ...
class ArrowTypeError(TypeError, ArrowException): ...
class ArrowNotImplementedError(NotImplementedError, ArrowException): ...
class ArrowCapacityError(ArrowException): ...
class ArrowIndexError(IndexError, ArrowException): ...
class ArrowSerializationError(ArrowException): ...

class ArrowCancelled(ArrowException):
    signum: int | None
    def __init__(self, message: str, signum: int | None = None) -> None: ...

ArrowIOError = IOError

class StopToken: ...

def enable_signal_handlers(enable: bool) -> None: ...

have_signal_refcycle: bool

class SignalStopHandler:
    def __enter__(self) -> Self: ...
    def __exit__(self, exc_type, exc_value, exc_tb) -> None: ...
    def __dealloc__(self) -> None: ...
    @property
    def stop_token(self) -> StopToken: ...

__all__ = [
    "ArrowException",
    "ArrowInvalid",
    "ArrowMemoryError",
    "ArrowKeyError",
    "ArrowTypeError",
    "ArrowNotImplementedError",
    "ArrowCapacityError",
    "ArrowIndexError",
    "ArrowSerializationError",
    "ArrowCancelled",
    "ArrowIOError",
    "StopToken",
    "enable_signal_handlers",
    "have_signal_refcycle",
    "SignalStopHandler",
]



from typing import Iterable, Literal

from .lib import Array, DataType, Field, MemoryPool, RecordBatch, Schema, _Weakrefable

class Node(_Weakrefable):
    def return_type(self) -> DataType: ...

class Expression(_Weakrefable):
    def root(self) -> Node: ...
    def result(self) -> Field: ...

class Condition(_Weakrefable):
    def root(self) -> Node: ...
    def result(self) -> Field: ...

class SelectionVector(_Weakrefable):
    def to_array(self) -> Array: ...

class Projector(_Weakrefable):
    @property
    def llvm_ir(self): ...
    def evaluate(
        self, batch: RecordBatch, selection: SelectionVector | None = None
    ) -> list[Array]: ...

class Filter(_Weakrefable):
    @property
    def llvm_ir(self): ...
    def evaluate(
        self, batch: RecordBatch, pool: MemoryPool, dtype: DataType | str = "int32"
    ) -> SelectionVector: ...

class TreeExprBuilder(_Weakrefable):
    def make_literal(self, value: float | str | bytes | bool, dtype: DataType) -> Node: ...
    def make_expression(self, root_node: Node, return_field: Field) -> Expression: ...
    def make_function(self, name: str, children: list[Node], return_type: DataType) -> Node: ...
    def make_field(self, field: Field) -> Node: ...
    def make_if(
        self, condition: Node, this_node: Node, else_node: Node, return_type: DataType
    ) -> Node: ...
    def make_and(self, children: list[Node]) -> Node: ...
    def make_or(self, children: list[Node]) -> Node: ...
    def make_in_expression(self, node: Node, values: Iterable, dtype: DataType) -> Node: ...
    def make_condition(self, condition: Node) -> Condition: ...

class Configuration(_Weakrefable):
    def __init__(self, optimize: bool = True, dump_ir: bool = False) -> None: ...

def make_projector(
    schema: Schema,
    children: list[Expression],
    pool: MemoryPool,
    selection_mode: Literal["NONE", "UINT16", "UINT32", "UINT64"] = "NONE",
    configuration: Configuration | None = None,
) -> Projector: ...
def make_filter(
    schema: Schema, condition: Condition, configuration: Configuration | None = None
) -> Filter: ...

class FunctionSignature(_Weakrefable):
    def return_type(self) -> DataType: ...
    def param_types(self) -> list[DataType]: ...
    def name(self) -> str: ...

def get_registered_function_signatures() -> list[FunctionSignature]: ...



# mypy: disable-error-code="overload-overlap,type-arg,misc"

import datetime as dt
import sys

from decimal import Decimal

if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self
if sys.version_info >= (3, 10):
    from typing import TypeAlias
else:
    from typing_extensions import TypeAlias
from typing import (
    Any,
    Generator,
    Generic,
    Iterable,
    Iterator,
    Literal,
    Mapping,
    Sequence,
    TypeVar,
    overload,
)

import numpy as np
import pandas as pd

from pyarrow._compute import CastOptions, FunctionOptions
from pyarrow._stubs_typing import (
    Indices,
    Mask,
    NullEncoding,
    NullSelectionBehavior,
    Order,
    SupportArrowArray,
    SupportArrowDeviceArray,
    SupportArrowStream,
)
from pyarrow.compute import Expression
from pyarrow.interchange.dataframe import _PyArrowDataFrame
from pyarrow.lib import Field, MemoryPool, MonthDayNano, Schema

from . import scalar
from .array import Array, NullableIterable, StructArray, _CastAs, _PandasConvertible
from .device import DeviceAllocationType
from .io import Buffer
from .ipc import RecordBatchReader
from .scalar import Int64Scalar, Scalar
from .tensor import Tensor
from .types import DataType, _AsPyType, _BasicDataType, _DataTypeT

_Scalar_CoT = TypeVar("_Scalar_CoT", bound=Scalar, covariant=True)

class ChunkedArray(_PandasConvertible[pd.Series], Generic[_Scalar_CoT]):
    @property
    def data(self) -> Self: ...
    @property
    def type(self) -> DataType: ...
    def length(self) -> int: ...
    __len__ = length
    def to_string(
        self,
        *,
        indent: int = 0,
        window: int = 5,
        container_window: int = 2,
        skip_new_lines: bool = False,
    ) -> str: ...
    format = to_string
    def validate(self, *, full: bool = False) -> None: ...
    @property
    def null_count(self) -> int: ...
    @property
    def nbytes(self) -> int: ...
    def get_total_buffer_size(self) -> int: ...
    def __sizeof__(self) -> int: ...
    @overload
    def __getitem__(self, key: slice) -> Self: ...
    @overload
    def __getitem__(self, key: int) -> _Scalar_CoT: ...
    def getitem(self, i: int) -> Scalar: ...
    def is_null(self, *, nan_is_null: bool = False) -> ChunkedArray[scalar.BooleanScalar]: ...
    def is_nan(self) -> ChunkedArray[scalar.BooleanScalar]: ...
    def is_valid(self) -> ChunkedArray[scalar.BooleanScalar]: ...
    def fill_null(self, fill_value: Scalar[_DataTypeT]) -> Self: ...
    def equals(self, other: Self) -> bool: ...
    def to_numpy(self, zero_copy_only: bool = False) -> np.ndarray: ...
    def __array__(self, dtype: np.dtype | None = None, copy: bool | None = None) -> np.ndarray: ...
    @overload
    def cast(
        self,
        target_type: None = None,
        safe: bool | None = None,
        options: CastOptions | None = None,
    ) -> Self: ...
    @overload
    def cast(
        self, target_type: _CastAs, safe: bool | None = None, options: CastOptions | None = None
    ) -> ChunkedArray[Scalar[_CastAs]]: ...
    def dictionary_encode(self, null_encoding: NullEncoding = "mask") -> Self: ...
    def flatten(self, memory_pool: MemoryPool | None = None) -> list[ChunkedArray[Any]]: ...
    def combine_chunks(self, memory_pool: MemoryPool | None = None) -> Array[_Scalar_CoT]: ...
    def unique(self) -> ChunkedArray[_Scalar_CoT]: ...
    def value_counts(self) -> StructArray: ...
    def slice(self, offset: int = 0, length: int | None = None) -> Self: ...
    def filter(self, mask: Mask, null_selection_behavior: NullSelectionBehavior = "drop"): ...
    @overload
    def index(
        self: ChunkedArray[Scalar[_BasicDataType[_AsPyType]]],
        value: Scalar[_DataTypeT] | _AsPyType,
        start: int | None = None,
        end: int | None = None,
        *,
        memory_pool: MemoryPool | None = None,
    ) -> Int64Scalar: ...
    @overload
    def index(
        self,
        value: Scalar[_DataTypeT],
        start: int | None = None,
        end: int | None = None,
        *,
        memory_pool: MemoryPool | None = None,
    ) -> Int64Scalar: ...
    def take(self, indices: Indices) -> Self: ...
    def drop_null(self) -> Self: ...
    def sort(self, order: Order = "ascending", **kwargs) -> Self: ...
    def unify_dictionaries(self, memory_pool: MemoryPool | None = None) -> Self: ...
    @property
    def num_chunks(self) -> int: ...
    def chunk(self, i: int) -> ChunkedArray[_Scalar_CoT]: ...
    @property
    def chunks(self) -> list[Array[_Scalar_CoT]]: ...
    def iterchunks(self) -> Generator[Array[_Scalar_CoT], None, None]: ...
    def __iter__(self) -> Iterator[Array[_Scalar_CoT]]: ...
    def to_pylist(
        self: ChunkedArray[Scalar[_BasicDataType[_AsPyType]]],
    ) -> list[_AsPyType | None]: ...
    def __arrow_c_stream__(self, requested_schema=None) -> Any: ...
    @classmethod
    def _import_from_c_capsule(cls, stream) -> Self: ...

@overload  # type: ignore[overload-overlap]
def chunked_array(
    values: NullableIterable[bool],
    type: None = None,
) -> ChunkedArray[scalar.BooleanScalar]: ...
@overload
def chunked_array(
    values: NullableIterable[int],
    type: None = None,
) -> ChunkedArray[scalar.Int64Scalar]: ...
@overload
def chunked_array(
    values: NullableIterable[float],
    type: None = None,
) -> ChunkedArray[scalar.DoubleScalar]: ...
@overload
def chunked_array(
    values: NullableIterable[Decimal],
    type: None = None,
) -> ChunkedArray[scalar.Decimal128Scalar]: ...
@overload
def chunked_array(
    values: NullableIterable[dict[str, Any]],
    type: None = None,
) -> ChunkedArray[scalar.StructScalar]: ...
@overload
def chunked_array(
    values: NullableIterable[dt.datetime],
    type: None = None,
) -> ChunkedArray[scalar.TimestampScalar]: ...
@overload
def chunked_array(
    values: NullableIterable[dt.date],
    type: None = None,
) -> ChunkedArray[scalar.Date32Scalar]: ...
@overload
def chunked_array(
    values: NullableIterable[dt.time],
    type: None = None,
) -> ChunkedArray[scalar.Time64Scalar]: ...
@overload
def chunked_array(
    values: NullableIterable[dt.timedelta],
    type: None = None,
) -> ChunkedArray[scalar.DurationScalar]: ...
@overload
def chunked_array(
    values: NullableIterable[MonthDayNano],
    type: None = None,
) -> ChunkedArray[scalar.MonthDayNanoIntervalScalar]: ...
@overload
def chunked_array(
    values: NullableIterable[str],
    type: None = None,
) -> ChunkedArray[scalar.StringScalar]: ...
@overload
def chunked_array(
    values: NullableIterable[bytearray],
    type: None = None,
) -> ChunkedArray[scalar.BinaryScalar]: ...
@overload
def chunked_array(
    values: NullableIterable[list],
    type: None = None,
) -> ChunkedArray[scalar.ListScalar]: ...
@overload
def chunked_array(
    values: NullableIterable[_Scalar_CoT],
    type: None = None,
) -> ChunkedArray[_Scalar_CoT]: ...
@overload
def chunked_array(
    values: Iterable | SupportArrowStream | SupportArrowArray,
    type: _DataTypeT,
) -> ChunkedArray[Scalar[_DataTypeT]]: ...
@overload
def chunked_array(
    values: Iterable | SupportArrowStream | SupportArrowArray,
    type: Literal["null"],
) -> ChunkedArray[scalar.NullScalar]: ...
@overload
def chunked_array(
    values: Iterable | SupportArrowStream | SupportArrowArray,
    type: Literal["bool", "boolean"],
) -> ChunkedArray[scalar.BooleanScalar]: ...
@overload
def chunked_array(
    values: Iterable | SupportArrowStream | SupportArrowArray,
    type: Literal["i1", "int8"],
) -> ChunkedArray[scalar.Int8Scalar]: ...
@overload
def chunked_array(
    values: Iterable | SupportArrowStream | SupportArrowArray,
    type: Literal["i2", "int16"],
) -> ChunkedArray[scalar.Int16Scalar]: ...
@overload
def chunked_array(
    values: Iterable | SupportArrowStream | SupportArrowArray,
    type: Literal["i4", "int32"],
) -> ChunkedArray[scalar.Int32Scalar]: ...
@overload
def chunked_array(
    values: Iterable | SupportArrowStream | SupportArrowArray,
    type: Literal["i8", "int64"],
) -> ChunkedArray[scalar.Int64Scalar]: ...
@overload
def chunked_array(
    values: Iterable | SupportArrowStream | SupportArrowArray,
    type: Literal["u1", "uint8"],
) -> ChunkedArray[scalar.UInt8Scalar]: ...
@overload
def chunked_array(
    values: Iterable | SupportArrowStream | SupportArrowArray,
    type: Literal["u2", "uint16"],
) -> ChunkedArray[scalar.UInt16Scalar]: ...
@overload
def chunked_array(
    values: Iterable | SupportArrowStream | SupportArrowArray,
    type: Literal["u4", "uint32"],
) -> ChunkedArray[scalar.UInt32Scalar]: ...
@overload
def chunked_array(
    values: Iterable | SupportArrowStream | SupportArrowArray,
    type: Literal["u8", "uint64"],
) -> ChunkedArray[scalar.UInt64Scalar]: ...
@overload
def chunked_array(
    values: Iterable | SupportArrowStream | SupportArrowArray,
    type: Literal["f2", "halffloat", "float16"],
) -> ChunkedArray[scalar.HalfFloatScalar]: ...
@overload
def chunked_array(
    values: Iterable | SupportArrowStream | SupportArrowArray,
    type: Literal["f4", "float", "float32"],
) -> ChunkedArray[scalar.FloatScalar]: ...
@overload
def chunked_array(
    values: Iterable | SupportArrowStream | SupportArrowArray,
    type: Literal["f8", "double", "float64"],
) -> ChunkedArray[scalar.DoubleScalar]: ...
@overload
def chunked_array(
    values: Iterable | SupportArrowStream | SupportArrowArray,
    type: Literal["string", "str", "utf8"],
) -> ChunkedArray[scalar.StringScalar]: ...
@overload
def chunked_array(
    values: Iterable | SupportArrowStream | SupportArrowArray,
    type: Literal["binary"],
) -> ChunkedArray[scalar.BinaryScalar]: ...
@overload
def chunked_array(
    values: Iterable | SupportArrowStream | SupportArrowArray,
    type: Literal["large_string", "large_str", "large_utf8"],
) -> ChunkedArray[scalar.LargeStringScalar]: ...
@overload
def chunked_array(
    values: Iterable | SupportArrowStream | SupportArrowArray,
    type: Literal["large_binary"],
) -> ChunkedArray[scalar.LargeBinaryScalar]: ...
@overload
def chunked_array(
    values: Iterable | SupportArrowStream | SupportArrowArray,
    type: Literal["binary_view"],
) -> ChunkedArray[scalar.BinaryViewScalar]: ...
@overload
def chunked_array(
    values: Iterable | SupportArrowStream | SupportArrowArray,
    type: Literal["string_view"],
) -> ChunkedArray[scalar.StringViewScalar]: ...
@overload
def chunked_array(
    values: Iterable | SupportArrowStream | SupportArrowArray,
    type: Literal["date32", "date32[day]"],
) -> ChunkedArray[scalar.Date32Scalar]: ...
@overload
def chunked_array(
    values: Iterable | SupportArrowStream | SupportArrowArray,
    type: Literal["date64", "date64[ms]"],
) -> ChunkedArray[scalar.Date64Scalar]: ...
@overload
def chunked_array(
    values: Iterable | SupportArrowStream | SupportArrowArray,
    type: Literal["time32[s]"],
) -> ChunkedArray[scalar.Time32Scalar[Literal["s"]]]: ...
@overload
def chunked_array(
    values: Iterable | SupportArrowStream | SupportArrowArray,
    type: Literal["time32[ms]"],
) -> ChunkedArray[scalar.Time32Scalar[Literal["ms"]]]: ...
@overload
def chunked_array(
    values: Iterable | SupportArrowStream | SupportArrowArray,
    type: Literal["time64[us]"],
) -> ChunkedArray[scalar.Time64Scalar[Literal["us"]]]: ...
@overload
def chunked_array(
    values: Iterable | SupportArrowStream | SupportArrowArray,
    type: Literal["time64[ns]"],
) -> ChunkedArray[scalar.Time64Scalar[Literal["ns"]]]: ...
@overload
def chunked_array(
    values: Iterable | SupportArrowStream | SupportArrowArray,
    type: Literal["timestamp[s]"],
) -> ChunkedArray[scalar.TimestampScalar[Literal["s"]]]: ...
@overload
def chunked_array(
    values: Iterable | SupportArrowStream | SupportArrowArray,
    type: Literal["timestamp[ms]"],
) -> ChunkedArray[scalar.TimestampScalar[Literal["ms"]]]: ...
@overload
def chunked_array(
    values: Iterable | SupportArrowStream | SupportArrowArray,
    type: Literal["timestamp[us]"],
) -> ChunkedArray[scalar.TimestampScalar[Literal["us"]]]: ...
@overload
def chunked_array(
    values: Iterable | SupportArrowStream | SupportArrowArray,
    type: Literal["timestamp[ns]"],
) -> ChunkedArray[scalar.TimestampScalar[Literal["ns"]]]: ...
@overload
def chunked_array(
    values: Iterable | SupportArrowStream | SupportArrowArray,
    type: Literal["duration[s]"],
) -> ChunkedArray[scalar.DurationScalar[Literal["s"]]]: ...
@overload
def chunked_array(
    values: Iterable | SupportArrowStream | SupportArrowArray,
    type: Literal["duration[ms]"],
) -> ChunkedArray[scalar.DurationScalar[Literal["ms"]]]: ...
@overload
def chunked_array(
    values: Iterable | SupportArrowStream | SupportArrowArray,
    type: Literal["duration[us]"],
) -> ChunkedArray[scalar.DurationScalar[Literal["us"]]]: ...
@overload
def chunked_array(
    values: Iterable | SupportArrowStream | SupportArrowArray,
    type: Literal["duration[ns]"],
) -> ChunkedArray[scalar.DurationScalar[Literal["ns"]]]: ...
@overload
def chunked_array(
    values: Iterable | SupportArrowStream | SupportArrowArray,
    type: Literal["month_day_nano_interval"],
) -> ChunkedArray[scalar.MonthDayNanoIntervalScalar]: ...

_ColumnT = TypeVar("_ColumnT", bound=Array | ChunkedArray)

class _Tabular(_PandasConvertible[pd.DataFrame], Generic[_ColumnT]):
    def __array__(self, dtype: np.dtype | None = None, copy: bool | None = None) -> np.ndarray: ...
    def __dataframe__(
        self, nan_as_null: bool = False, allow_copy: bool = True
    ) -> _PyArrowDataFrame: ...
    @overload
    def __getitem__(self, key: int | str) -> _ColumnT: ...
    @overload
    def __getitem__(self, key: slice) -> Self: ...
    def __len__(self) -> int: ...
    def column(self, i: int | str) -> _ColumnT: ...
    @property
    def column_names(self) -> list[str]: ...
    @property
    def columns(self) -> list[_ColumnT]: ...
    def drop_null(self) -> Self: ...
    def field(self, i: int | str) -> Field: ...
    @classmethod
    def from_pydict(
        cls,
        mapping: Mapping[str, ChunkedArray | Array | list | np.ndarray],
        schema: Schema | None = None,
        metadata: Mapping | None = None,
    ) -> Self: ...
    @classmethod
    def from_pylist(
        cls,
        mapping: Sequence[Mapping[str, Any]],
        schema: Schema | None = None,
        metadata: Mapping | None = None,
    ) -> Self: ...
    def itercolumns(self) -> Generator[_ColumnT, None, None]: ...
    @property
    def num_columns(self) -> int: ...
    @property
    def num_rows(self) -> int: ...
    @property
    def shape(self) -> tuple[int, int]: ...
    @property
    def schema(self) -> Schema: ...
    @property
    def nbytes(self) -> int: ...
    def sort_by(self, sorting: str | list[tuple[str, Order]], **kwargs) -> Self: ...
    def take(self, indices: Indices) -> Self: ...
    def filter(
        self, mask: Mask | Expression, null_selection_behavior: NullSelectionBehavior = "drop"
    ) -> Self: ...
    def to_pydict(self) -> dict[str, list]: ...
    def to_pylist(self) -> list[dict[str, Any]]: ...
    def to_string(self, *, show_metadata: bool = False, preview_cols: int = 0) -> str: ...
    def remove_column(self, i: int) -> Self: ...
    def drop_columns(self, columns: str | list[str]) -> Self: ...
    def add_column(
        self, i: int, field_: str | Field, column: ChunkedArray | Array | list
    ) -> Self: ...
    def append_column(self, field_: str | Field, column: ChunkedArray | Array | list) -> Self: ...

class RecordBatch(_Tabular[Array]):
    def validate(self, *, full: bool = False) -> None: ...
    def replace_schema_metadata(self, metadata: dict | None = None) -> Self: ...
    def get_total_buffer_size(self) -> int: ...
    def __sizeof__(self) -> int: ...
    def set_column(self, i: int, field_: str | Field, column: Array | list) -> Self: ...
    @overload
    def rename_columns(self, names: list[str]) -> Self: ...
    @overload
    def rename_columns(self, names: dict[str, str]) -> Self: ...
    def serialize(self, memory_pool: MemoryPool | None = None) -> Buffer: ...
    def slice(self, offset: int = 0, length: int | None = None) -> Self: ...
    def equals(self, other: Self, check_metadata: bool = False) -> bool: ...
    def select(self, columns: list[str] | Indices) -> Self: ...
    def cast(
        self, target_schema: Schema, safe: bool | None = None, options: CastOptions | None = None
    ) -> Self: ...
    @classmethod
    def from_arrays(
        cls,
        arrays: list[Array] | list[ChunkedArray],
        names: list[str] | None = None,
        schema: Schema | None = None,
        metadata: Mapping | None = None,
    ) -> Self: ...
    @classmethod
    def from_pandas(
        cls,
        df: pd.DataFrame,
        schema: Schema | None = None,
        preserve_index: bool | None = None,
        nthreads: int | None = None,
        columns: list[str] | None = None,
    ) -> Self: ...
    @classmethod
    def from_struct_array(
        cls, struct_array: StructArray | ChunkedArray[scalar.StructScalar]
    ) -> Self: ...
    def to_struct_array(self) -> StructArray: ...
    def to_tensor(
        self,
        null_to_nan: bool = False,
        row_major: bool = True,
        memory_pool: MemoryPool | None = None,
    ) -> Tensor: ...
    def _export_to_c(self, out_ptr: int, out_schema_ptr: int = 0): ...
    @classmethod
    def _import_from_c(cls, in_ptr: int, schema: Schema) -> Self: ...
    def __arrow_c_array__(self, requested_schema=None): ...
    def __arrow_c_stream__(self, requested_schema=None): ...
    @classmethod
    def _import_from_c_capsule(cls, schema_capsule, array_capsule) -> Self: ...
    def _export_to_c_device(self, out_ptr: int, out_schema_ptr: int = 0) -> None: ...
    @classmethod
    def _import_from_c_device(cls, in_ptr: int, schema: Schema) -> Self: ...
    def __arrow_c_device_array__(self, requested_schema=None, **kwargs): ...
    @classmethod
    def _import_from_c_device_capsule(cls, schema_capsule, array_capsule) -> Self: ...
    @property
    def device_type(self) -> DeviceAllocationType: ...
    @property
    def is_cpu(self) -> bool: ...

def table_to_blocks(options, table: Table, categories, extension_columns): ...

JoinType: TypeAlias = Literal[
    "left semi",
    "right semi",
    "left anti",
    "right anti",
    "inner",
    "left outer",
    "right outer",
    "full outer",
]

class Table(_Tabular[ChunkedArray]):
    def validate(self, *, full=False) -> None: ...
    def slice(self, offset=0, length=None) -> Self: ...
    def select(self, columns: list[str] | Indices) -> Self: ...
    def replace_schema_metadata(self, metadata: dict | None = None) -> Self: ...
    def flatten(self, memory_pool: MemoryPool | None = None) -> Self: ...
    def combine_chunks(self, memory_pool: MemoryPool | None = None) -> Self: ...
    def unify_dictionaries(self, memory_pool: MemoryPool | None = None) -> Self: ...
    def equals(self, other: Self, check_metadata: bool = False) -> Self: ...
    def cast(
        self, target_schema: Schema, safe: bool | None = None, options: CastOptions | None = None
    ) -> Self: ...
    @classmethod
    def from_pandas(
        cls,
        df: pd.DataFrame,
        schema: Schema | None = None,
        preserve_index: bool | None = None,
        nthreads: int | None = None,
        columns: list[str] | None = None,
        safe: bool = True,
    ) -> Self: ...
    @classmethod
    def from_arrays(
        cls,
        arrays: list[Array] | list[ChunkedArray],
        names: list[str] | None = None,
        schema: Schema | None = None,
        metadata: Mapping | None = None,
    ) -> Self: ...
    @classmethod
    def from_struct_array(
        cls, struct_array: StructArray | ChunkedArray[scalar.StructScalar]
    ) -> Self: ...
    def to_struct_array(
        self, max_chunksize: int | None = None
    ) -> ChunkedArray[scalar.StructScalar]: ...
    @classmethod
    def from_batches(
        cls, batches: Iterable[RecordBatch], schema: Schema | None = None
    ) -> Self: ...
    def to_batches(self, max_chunksize: int | None = None) -> list[RecordBatch]: ...
    def to_reader(self, max_chunksize: int | None = None) -> RecordBatchReader: ...
    def get_total_buffer_size(self) -> int: ...
    def __sizeof__(self) -> int: ...
    def set_column(self, i: int, field_: str | Field, column: Array | list) -> Self: ...
    @overload
    def rename_columns(self, names: list[str]) -> Self: ...
    @overload
    def rename_columns(self, names: dict[str, str]) -> Self: ...
    def drop(self, columns: str | list[str]) -> Self: ...
    def group_by(self, keys: str | list[str], use_threads: bool = True) -> TableGroupBy: ...
    def join(
        self,
        right_table: Self,
        keys: str | list[str],
        right_keys: str | list[str] | None = None,
        join_type: JoinType = "left outer",
        left_suffix: str | None = None,
        right_suffix: str | None = None,
        coalesce_keys: bool = True,
        use_threads: bool = True,
    ) -> Self: ...
    def join_asof(
        self,
        right_table: Self,
        on: str,
        by: str | list[str],
        tolerance: int,
        right_on: str | list[str] | None = None,
        right_by: str | list[str] | None = None,
    ) -> Self: ...
    def __arrow_c_stream__(self, requested_schema=None): ...

def record_batch(
    data: dict[str, list | Array]
    | list[Array]
    | pd.DataFrame
    | SupportArrowArray
    | SupportArrowDeviceArray,
    names: list[str] | None = None,
    schema: Schema | None = None,
    metadata: Mapping | None = None,
) -> RecordBatch: ...
@overload
def table(
    data: dict[str, list | Array],
    schema: Schema | None = None,
    metadata: Mapping | None = None,
    nthreads: int | None = None,
) -> Table: ...
@overload
def table(
    data: list[Array | ChunkedArray]
    | pd.DataFrame
    | SupportArrowArray
    | SupportArrowStream
    | SupportArrowDeviceArray,
    names: list[str] | None = None,
    schema: Schema | None = None,
    metadata: Mapping | None = None,
    nthreads: int | None = None,
) -> Table: ...
def concat_tables(
    tables: list[Table],
    memory_pool: MemoryPool | None = None,
    promote_options: Literal["none", "default", "permissive"] = "none",
    **kwargs,
) -> Table: ...

class TableGroupBy:
    keys: str | list[str]
    def __init__(self, table: Table, keys: str | list[str], use_threads: bool = True): ...
    def aggregate(
        self, aggregations: list[tuple[str, str]] | list[tuple[str, str, FunctionOptions]]
    ) -> Table: ...
    def _table(self) -> Table: ...
    @property
    def _use_threads(self) -> bool: ...

__all__ = [
    "ChunkedArray",
    "chunked_array",
    "_Tabular",
    "RecordBatch",
    "table_to_blocks",
    "Table",
    "record_batch",
    "table",
    "concat_tables",
    "TableGroupBy",
]



from typing import Any, Protocol, TypeAlias

from pyarrow.lib import Array, Buffer, DataType, DictionaryArray, RecordBatch, Table

from .column import (
    ColumnBuffers,
    ColumnNullType,
    Dtype,
    DtypeKind,
)

class DataFrameObject(Protocol):
    def __dataframe__(self, nan_as_null: bool = False, allow_copy: bool = True) -> Any: ...

ColumnObject: TypeAlias = Any

def from_dataframe(df: DataFrameObject, allow_copy=True) -> Table: ...
def protocol_df_chunk_to_pyarrow(df: DataFrameObject, allow_copy: bool = True) -> RecordBatch: ...
def column_to_array(col: ColumnObject, allow_copy: bool = True) -> Array: ...
def bool_column_to_array(col: ColumnObject, allow_copy: bool = True) -> Array: ...
def categorical_column_to_dictionary(
    col: ColumnObject, allow_copy: bool = True
) -> DictionaryArray: ...
def parse_datetime_format_str(format_str: str) -> tuple[str, str]: ...
def map_date_type(data_type: tuple[DtypeKind, int, str, str]) -> DataType: ...
def buffers_to_array(
    buffers: ColumnBuffers,
    data_type: tuple[DtypeKind, int, str, str],
    length: int,
    describe_null: ColumnNullType,
    offset: int = 0,
    allow_copy: bool = True,
) -> Array: ...
def validity_buffer_from_mask(
    validity_buff: Buffer,
    validity_dtype: Dtype,
    describe_null: ColumnNullType,
    length: int,
    offset: int = 0,
    allow_copy: bool = True,
) -> Buffer: ...
def validity_buffer_nan_sentinel(
    data_pa_buffer: Buffer,
    data_type: Dtype,
    describe_null: ColumnNullType,
    length: int,
    offset: int = 0,
    allow_copy: bool = True,
) -> Buffer: ...



import enum
import sys

from io import IOBase

if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self
from typing import Iterable, Iterator, Literal, Mapping, NamedTuple

import pandas as pd

from pyarrow._stubs_typing import SupportArrowStream, SupportPyBuffer
from pyarrow.lib import MemoryPool, RecordBatch, Schema, Table, Tensor, _Weakrefable

from .io import Buffer, Codec, NativeFile
from .types import DictionaryMemo, KeyValueMetadata

class MetadataVersion(enum.IntEnum):
    V1 = enum.auto()
    V2 = enum.auto()
    V3 = enum.auto()
    V4 = enum.auto()
    V5 = enum.auto()

class WriteStats(NamedTuple):
    num_messages: int
    num_record_batches: int
    num_dictionary_batches: int
    num_dictionary_deltas: int
    num_replaced_dictionaries: int

class ReadStats(NamedTuple):
    num_messages: int
    num_record_batches: int
    num_dictionary_batches: int
    num_dictionary_deltas: int
    num_replaced_dictionaries: int

class IpcReadOptions(_Weakrefable):
    ensure_native_endian: bool
    use_threads: bool
    include_fields: list
    def __init__(
        self, *, ensure_native_endian: bool = True, use_threads: bool = True, include_fields: list
    ) -> None: ...

class IpcWriteOptions(_Weakrefable):
    metadata_version: MetadataVersion
    allow_64bit: bool
    use_legacy_format: bool
    compression: Codec | Literal["lz4", "zstd"] | None
    use_threads: bool
    emit_dictionary_deltas: bool
    unify_dictionaries: bool
    def __init__(
        self,
        *,
        metadata_version: MetadataVersion = MetadataVersion.V5,
        allow_64bit: bool = False,
        use_legacy_format: bool = False,
        compression: Codec | Literal["lz4", "zstd"] | None = None,
        use_threads: bool = True,
        emit_dictionary_deltas: bool = False,
        unify_dictionaries: bool = False,
    ) -> None: ...

class Message(_Weakrefable):
    @property
    def type(self) -> str: ...
    @property
    def metadata(self) -> Buffer: ...
    @property
    def metadata_version(self) -> MetadataVersion: ...
    @property
    def body(self) -> Buffer | None: ...
    def equals(self, other: Message) -> bool: ...
    def serialize_to(
        self, sink: NativeFile, alignment: int = 8, memory_pool: MemoryPool | None = None
    ): ...
    def serialize(self, alignment: int = 8, memory_pool: MemoryPool | None = None) -> Buffer: ...

class MessageReader(_Weakrefable):
    @classmethod
    def open_stream(cls, source: bytes | NativeFile | IOBase | SupportPyBuffer) -> Self: ...
    def __iter__(self) -> Self: ...
    def read_next_message(self) -> Message: ...
    __next__ = read_next_message

# ----------------------------------------------------------------------
# File and stream readers and writers

class _CRecordBatchWriter(_Weakrefable):
    def write(self, table_or_batch: Table | RecordBatch): ...
    def write_batch(
        self,
        batch: RecordBatch,
        custom_metadata: Mapping[bytes, bytes] | KeyValueMetadata | None = None,
    ): ...
    def write_table(self, table: Table, max_chunksize: int | None = None) -> None: ...
    def close(self) -> None: ...
    def __enter__(self) -> Self: ...
    def __exit__(self, exc_type, exc_val, exc_tb): ...
    @property
    def stats(self) -> WriteStats: ...

class _RecordBatchStreamWriter(_CRecordBatchWriter):
    def __dealloc__(self) -> None: ...
    def _open(self, sink, schema: Schema, options: IpcWriteOptions = IpcWriteOptions()): ...

class _ReadPandasMixin:
    def read_pandas(self, **options) -> pd.DataFrame: ...

class RecordBatchReader(_Weakrefable):
    def __iter__(self) -> Self: ...
    def read_next_batch(self) -> RecordBatch: ...
    __next__ = read_next_batch
    @property
    def schema(self) -> Schema: ...
    def read_next_batch_with_custom_metadata(self) -> RecordBatchWithMetadata: ...
    def iter_batches_with_custom_metadata(
        self,
    ) -> Iterator[RecordBatchWithMetadata]: ...
    def read_all(self) -> Table: ...
    read_pandas = _ReadPandasMixin.read_pandas
    def close(self) -> None: ...
    def __enter__(self) -> Self: ...
    def __exit__(self, exc_type, exc_val, exc_tb): ...
    def cast(self, target_schema: Schema) -> Self: ...
    def _export_to_c(self, out_ptr: int) -> None: ...
    @classmethod
    def _import_from_c(cls, in_ptr: int) -> Self: ...
    def __arrow_c_stream__(self, requested_schema=None): ...
    @classmethod
    def _import_from_c_capsule(cls, stream) -> Self: ...
    @classmethod
    def from_stream(cls, data: SupportArrowStream, schema: Schema | None = None) -> Self: ...
    @classmethod
    def from_batches(cls, schema: Schema, batches: Iterable[RecordBatch]) -> Self: ...

class _RecordBatchStreamReader(RecordBatchReader):
    @property
    def stats(self) -> ReadStats: ...

class _RecordBatchFileWriter(_RecordBatchStreamWriter): ...

class RecordBatchWithMetadata(NamedTuple):
    batch: RecordBatch
    custom_metadata: KeyValueMetadata

class _RecordBatchFileReader(_Weakrefable):
    @property
    def num_record_batches(self) -> int: ...
    def get_batch(self, i: int) -> RecordBatch: ...
    get_record_batch = get_batch
    def get_batch_with_custom_metadata(self, i: int) -> RecordBatchWithMetadata: ...
    def read_all(self) -> Table: ...
    read_pandas = _ReadPandasMixin.read_pandas
    def __enter__(self) -> Self: ...
    def __exit__(self, exc_type, exc_val, exc_tb): ...
    @property
    def schema(self) -> Schema: ...
    @property
    def stats(self) -> ReadStats: ...

def get_tensor_size(tensor: Tensor) -> int: ...
def get_record_batch_size(batch: RecordBatch) -> int: ...
def write_tensor(tensor: Tensor, dest: NativeFile) -> int: ...
def read_tensor(source: NativeFile) -> Tensor: ...
def read_message(source: NativeFile | IOBase | SupportPyBuffer) -> Message: ...
def read_schema(
    obj: Buffer | Message, dictionary_memo: DictionaryMemo | None = None
) -> Schema: ...
def read_record_batch(
    obj: Message | SupportPyBuffer, schema: Schema, dictionary_memo: DictionaryMemo | None = None
) -> RecordBatch: ...

__all__ = [
    "MetadataVersion",
    "WriteStats",
    "ReadStats",
    "IpcReadOptions",
    "IpcWriteOptions",
    "Message",
    "MessageReader",
    "_CRecordBatchWriter",
    "_RecordBatchStreamWriter",
    "_ReadPandasMixin",
    "RecordBatchReader",
    "_RecordBatchStreamReader",
    "_RecordBatchFileWriter",
    "RecordBatchWithMetadata",
    "_RecordBatchFileReader",
    "get_tensor_size",
    "get_record_batch_size",
    "write_tensor",
    "read_tensor",
    "read_message",
    "read_schema",
    "read_record_batch",
]



from typing import IO, Literal

from .lib import (
    Buffer,
    KeyValueMetadata,
    MemoryPool,
    NativeFile,
    RecordBatch,
    Schema,
    Table,
    _Weakrefable,
)

class ORCReader(_Weakrefable):
    def __init__(self, memory_pool: MemoryPool | None = None) -> None: ...
    def open(self, source: str | NativeFile | Buffer, use_memory_map: bool = True): ...
    def metadata(self) -> KeyValueMetadata: ...
    def schema(self) -> Schema: ...
    def nrows(self) -> int: ...
    def nstripes(self) -> int: ...
    def file_version(self) -> str: ...
    def software_version(self) -> str: ...
    def compression(self) -> Literal["UNCOMPRESSED", "ZLIB", "SNAPPY", "LZ4", "ZSTD"]: ...
    def compression_size(self) -> int: ...
    def row_index_stride(self) -> int: ...
    def writer(self) -> str: ...
    def writer_version(self) -> str: ...
    def nstripe_statistics(self) -> int: ...
    def content_length(self) -> int: ...
    def stripe_statistics_length(self) -> int: ...
    def file_footer_length(self) -> int: ...
    def file_postscript_length(self) -> int: ...
    def file_length(self) -> int: ...
    def serialized_file_tail(self) -> int: ...
    def read_stripe(self, n: int, columns: list[str] | None = None) -> RecordBatch: ...
    def read(self, columns: list[str] | None = None) -> Table: ...

class ORCWriter(_Weakrefable):
    def open(
        self,
        where: str | NativeFile | IO,
        *,
        file_version: str | None = None,
        batch_size: int | None = None,
        stripe_size: int | None = None,
        compression: Literal["UNCOMPRESSED", "ZLIB", "SNAPPY", "LZ4", "ZSTD"] | None = None,
        compression_block_size: int | None = None,
        compression_strategy: Literal["COMPRESSION", "SPEED"] | None = None,
        row_index_stride: int | None = None,
        padding_tolerance: float | None = None,
        dictionary_key_size_threshold: float | None = None,
        bloom_filter_columns: list[int] | None = None,
        bloom_filter_fpp: float | None = None,
    ) -> None: ...
    def write(self, table: Table) -> None: ...
    def close(self) -> None: ...



import asyncio
import enum
import sys

if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self
from typing import Generator, Generic, Iterable, Iterator, NamedTuple, TypeVar

from typing_extensions import deprecated

from .ipc import _ReadPandasMixin
from .lib import (
    ArrowCancelled,
    ArrowException,
    ArrowInvalid,
    Buffer,
    IpcReadOptions,
    IpcWriteOptions,
    RecordBatch,
    RecordBatchReader,
    Schema,
    Table,
    _CRecordBatchWriter,
    _Weakrefable,
)

_T = TypeVar("_T")

class FlightCallOptions(_Weakrefable):
    def __init__(
        self,
        timeout: float | None = None,
        write_options: IpcWriteOptions | None = None,
        headers: list[tuple[str, str]] | None = None,
        read_options: IpcReadOptions | None = None,
    ) -> None: ...

class CertKeyPair(NamedTuple):
    cert: str
    key: str

class FlightError(Exception):
    extra_info: bytes

class FlightInternalError(FlightError, ArrowException): ...
class FlightTimedOutError(FlightError, ArrowException): ...
class FlightCancelledError(FlightError, ArrowCancelled): ...
class FlightServerError(FlightError, ArrowException): ...
class FlightUnauthenticatedError(FlightError, ArrowException): ...
class FlightUnauthorizedError(FlightError, ArrowException): ...
class FlightUnavailableError(FlightError, ArrowException): ...

class FlightWriteSizeExceededError(ArrowInvalid):
    limit: int
    actual: int

class Action(_Weakrefable):
    def __init__(self, action_type: bytes | str, buf: Buffer | bytes) -> None: ...
    @property
    def type(self) -> str: ...
    @property
    def body(self) -> Buffer: ...
    def serialize(self) -> bytes: ...
    @classmethod
    def deserialize(cls, serialized: bytes) -> Self: ...

class ActionType(NamedTuple):
    type: str
    description: str

    def make_action(self, buf: Buffer | bytes) -> Action: ...

class Result(_Weakrefable):
    def __init__(self, buf: Buffer | bytes) -> None: ...
    @property
    def body(self) -> Buffer: ...
    def serialize(self) -> bytes: ...
    @classmethod
    def deserialize(cls, serialized: bytes) -> Self: ...

class BasicAuth(_Weakrefable):
    def __init__(
        self, username: str | bytes | None = None, password: str | bytes | None = None
    ) -> None: ...
    @property
    def username(self) -> bytes: ...
    @property
    def password(self) -> bytes: ...
    def serialize(self) -> str: ...
    @staticmethod
    def deserialize(serialized: str | bytes) -> BasicAuth: ...

class DescriptorType(enum.Enum):
    UNKNOWN = 0
    PATH = 1
    CMD = 2

class FlightMethod(enum.Enum):
    INVALID = 0
    HANDSHAKE = 1
    LIST_FLIGHTS = 2
    GET_FLIGHT_INFO = 3
    GET_SCHEMA = 4
    DO_GET = 5
    DO_PUT = 6
    DO_ACTION = 7
    LIST_ACTIONS = 8
    DO_EXCHANGE = 9

class FlightDescriptor(_Weakrefable):
    @staticmethod
    def for_path(*path: str | bytes) -> FlightDescriptor: ...
    @staticmethod
    def for_command(command: str | bytes) -> FlightDescriptor: ...
    @property
    def descriptor_type(self) -> DescriptorType: ...
    @property
    def path(self) -> list[bytes] | None: ...
    @property
    def command(self) -> bytes | None: ...
    def serialize(self) -> bytes: ...
    @classmethod
    def deserialize(cls, serialized: bytes) -> Self: ...

class Ticket(_Weakrefable):
    def __init__(self, ticket: str | bytes) -> None: ...
    @property
    def ticket(self) -> bytes: ...
    def serialize(self) -> bytes: ...
    @classmethod
    def deserialize(cls, serialized: bytes) -> Self: ...

class Location(_Weakrefable):
    def __init__(self, uri: str | bytes) -> None: ...
    @property
    def uri(self) -> bytes: ...
    def equals(self, other: Location) -> bool: ...
    @staticmethod
    def for_grpc_tcp(host: str | bytes, port: int) -> Location: ...
    @staticmethod
    def for_grpc_tls(host: str | bytes, port: int) -> Location: ...
    @staticmethod
    def for_grpc_unix(path: str | bytes) -> Location: ...

class FlightEndpoint(_Weakrefable):
    def __init__(self, ticket: Ticket | str | bytes, locations: list[str | Location]): ...
    @property
    def ticket(self) -> Ticket: ...
    @property
    def locations(self) -> list[Location]: ...
    def serialize(self) -> bytes: ...
    @classmethod
    def deserialize(cls, serialized: bytes) -> Self: ...

class SchemaResult(_Weakrefable):
    def __init__(self, schema: Schema) -> None: ...
    @property
    def schema(self) -> Schema: ...
    def serialize(self) -> bytes: ...
    @classmethod
    def deserialize(cls, serialized: bytes) -> Self: ...

class FlightInfo(_Weakrefable):
    def __init__(
        self,
        schema: Schema,
        descriptor: FlightDescriptor,
        endpoints: list[FlightEndpoint],
        total_records: int,
        total_bytes: int,
    ) -> None: ...
    @property
    def schema(self) -> Schema: ...
    @property
    def descriptor(self) -> FlightDescriptor: ...
    @property
    def endpoints(self) -> list[FlightEndpoint]: ...
    @property
    def total_records(self) -> int: ...
    @property
    def total_bytes(self) -> int: ...
    def serialize(self) -> bytes: ...
    @classmethod
    def deserialize(cls, serialized: bytes) -> Self: ...

class FlightStreamChunk(_Weakrefable):
    @property
    def data(self) -> RecordBatch | None: ...
    @property
    def app_metadata(self) -> Buffer | None: ...
    def __iter__(self): ...

class _MetadataRecordBatchReader(_Weakrefable, _ReadPandasMixin):
    def __iter__(self) -> Self: ...
    def __next__(self) -> FlightStreamChunk: ...
    @property
    def schema(self) -> Schema: ...
    def read_all(self) -> Table: ...
    def read_chunk(self) -> FlightStreamChunk: ...
    def to_reader(self) -> RecordBatchReader: ...

class MetadataRecordBatchReader(_MetadataRecordBatchReader): ...

class FlightStreamReader(MetadataRecordBatchReader):
    def cancel(self) -> None: ...
    def read_all(self) -> Table: ...

class MetadataRecordBatchWriter(_CRecordBatchWriter):
    def begin(self, schema: Schema, options: IpcWriteOptions | None = None) -> None: ...
    def write_metadata(self, buf: Buffer) -> None: ...
    def write_batch(self, batch: RecordBatch) -> None: ...  # type: ignore[override]
    def write_table(self, table: Table, max_chunksize: int | None = None, **kwargs) -> None: ...
    def close(self) -> None: ...
    def write_with_metadata(self, batch: RecordBatch, buf: Buffer) -> None: ...

class FlightStreamWriter(MetadataRecordBatchWriter):
    def done_writing(self) -> None: ...

class FlightMetadataReader(_Weakrefable):
    def read(self) -> Buffer | None: ...

class FlightMetadataWriter(_Weakrefable):
    def write(self, message: Buffer) -> None: ...

class AsyncioCall(Generic[_T]):
    _future: asyncio.Future[_T]

    def as_awaitable(self) -> asyncio.Future[_T]: ...
    def wakeup(self, result_or_exception: BaseException | _T) -> None: ...

class AsyncioFlightClient:
    def __init__(self, client: FlightClient) -> None: ...
    async def get_flight_info(
        self,
        descriptor: FlightDescriptor,
        *,
        options: FlightCallOptions | None = None,
    ): ...

class FlightClient(_Weakrefable):
    def __init__(
        self,
        location: str | tuple[str, int] | Location,
        *,
        tls_root_certs: str | None = None,
        cert_chain: str | None = None,
        private_key: str | None = None,
        override_hostname: str | None = None,
        middleware: list[ClientMiddlewareFactory] | None = None,
        write_size_limit_bytes: int | None = None,
        disable_server_verification: bool = False,
        generic_options: list[tuple[str, int | str]] | None = None,
    ): ...
    @property
    def supports_async(self) -> bool: ...
    def as_async(self) -> AsyncioFlightClient: ...
    def wait_for_available(self, timeout: int = 5) -> None: ...
    @deprecated(
        "Use the ``FlightClient`` constructor or ``pyarrow.flight.connect`` function instead."
    )
    @classmethod
    def connect(
        cls,
        location: str | tuple[str, int] | Location,
        tls_root_certs: str | None = None,
        cert_chain: str | None = None,
        private_key: str | None = None,
        override_hostname: str | None = None,
        disable_server_verification: bool = False,
    ) -> FlightClient: ...
    def authenticate(
        self, auth_handler: ClientAuthHandler, options: FlightCallOptions | None = None
    ) -> None: ...
    def authenticate_basic_token(
        self, username: str, password: str, options: FlightCallOptions | None = None
    ) -> tuple[str, str]: ...
    def list_actions(self, options: FlightCallOptions | None = None) -> list[Action]: ...
    def do_action(
        self, action: Action, options: FlightCallOptions | None = None
    ) -> Iterator[Result]: ...
    def list_flights(
        self, criteria: str | None = None, options: FlightCallOptions | None = None
    ) -> Generator[FlightInfo, None, None]: ...
    def get_flight_info(
        self, descriptor: FlightDescriptor, options: FlightCallOptions | None = None
    ) -> FlightInfo: ...
    def get_schema(
        self, descriptor: FlightDescriptor, options: FlightCallOptions | None = None
    ) -> Schema: ...
    def do_get(
        self, ticket: Ticket, options: FlightCallOptions | None = None
    ) -> FlightStreamReader: ...
    def do_put(
        self,
        descriptor: FlightDescriptor,
        schema: Schema,
        options: FlightCallOptions | None = None,
    ) -> tuple[FlightStreamWriter, FlightStreamReader]: ...
    def do_exchange(
        self, descriptor: FlightDescriptor, options: FlightCallOptions | None = None
    ) -> tuple[FlightStreamWriter, FlightStreamReader]: ...
    def close(self) -> None: ...
    def __enter__(self) -> Self: ...
    def __exit__(self, exc_type, exc_value, traceback) -> None: ...

class FlightDataStream(_Weakrefable): ...

class RecordBatchStream(FlightDataStream):
    def __init__(
        self, data_source: RecordBatchReader | Table, options: IpcWriteOptions | None = None
    ) -> None: ...

class GeneratorStream(FlightDataStream):
    def __init__(
        self,
        schema: Schema,
        generator: Iterable[FlightDataStream | Table | RecordBatch | RecordBatchReader],
        options: IpcWriteOptions | None = None,
    ) -> None: ...

class ServerCallContext(_Weakrefable):
    def peer_identity(self) -> bytes: ...
    def peer(self) -> str: ...
    def is_cancelled(self) -> bool: ...
    def add_header(self, key: str, value: str) -> None: ...
    def add_trailer(self, key: str, value: str) -> None: ...
    def get_middleware(self, key: str) -> ServerMiddleware | None: ...

class ServerAuthReader(_Weakrefable):
    def read(self) -> str: ...

class ServerAuthSender(_Weakrefable):
    def write(self, message: str) -> None: ...

class ClientAuthReader(_Weakrefable):
    def read(self) -> str: ...

class ClientAuthSender(_Weakrefable):
    def write(self, message: str) -> None: ...

class ServerAuthHandler(_Weakrefable):
    def authenticate(self, outgoing: ServerAuthSender, incoming: ServerAuthReader): ...
    def is_valid(self, token: str) -> bool: ...

class ClientAuthHandler(_Weakrefable):
    def authenticate(self, outgoing: ClientAuthSender, incoming: ClientAuthReader): ...
    def get_token(self) -> str: ...

class CallInfo(NamedTuple):
    method: FlightMethod

class ClientMiddlewareFactory(_Weakrefable):
    def start_call(self, info: CallInfo) -> ClientMiddleware | None: ...

class ClientMiddleware(_Weakrefable):
    def sending_headers(self) -> dict[str, list[str] | list[bytes]]: ...
    def received_headers(self, headers: dict[str, list[str] | list[bytes]]): ...
    def call_completed(self, exception: ArrowException): ...

class ServerMiddlewareFactory(_Weakrefable):
    def start_call(
        self, info: CallInfo, headers: dict[str, list[str] | list[bytes]]
    ) -> ServerMiddleware | None: ...

class TracingServerMiddlewareFactory(ServerMiddlewareFactory): ...

class ServerMiddleware(_Weakrefable):
    def sending_headers(self) -> dict[str, list[str] | list[bytes]]: ...
    def call_completed(self, exception: ArrowException): ...

class TracingServerMiddleware(ServerMiddleware):
    trace_context: dict
    def __init__(self, trace_context: dict) -> None: ...

class _ServerMiddlewareFactoryWrapper(ServerMiddlewareFactory):
    def __init__(self, factories: dict[str, ServerMiddlewareFactory]) -> None: ...
    def start_call(  # type: ignore[override]
        self, info: CallInfo, headers: dict[str, list[str] | list[bytes]]
    ) -> _ServerMiddlewareFactoryWrapper | None: ...

class _ServerMiddlewareWrapper(ServerMiddleware):
    def __init__(self, middleware: dict[str, ServerMiddleware]) -> None: ...
    def send_headers(self) -> dict[str, dict[str, list[str] | list[bytes]]]: ...
    def call_completed(self, exception: ArrowException) -> None: ...

class _FlightServerFinalizer(_Weakrefable):
    def finalize(self) -> None: ...

class FlightServerBase(_Weakrefable):
    def __init__(
        self,
        location: str | tuple[str, int] | Location | None = None,
        auth_handler: ServerAuthHandler | None = None,
        tls_certificates: list[tuple[str, str]] | None = None,
        verify_client: bool = False,
        root_certificates: str | None = None,
        middleware: dict[str, ServerMiddlewareFactory] | None = None,
    ): ...
    @property
    def port(self) -> int: ...
    def list_flights(self, context: ServerCallContext, criteria: str) -> Iterator[FlightInfo]: ...
    def get_flight_info(
        self, context: ServerCallContext, descriptor: FlightDescriptor
    ) -> FlightInfo: ...
    def get_schema(self, context: ServerCallContext, descriptor: FlightDescriptor) -> Schema: ...
    def do_put(
        self,
        context: ServerCallContext,
        descriptor: FlightDescriptor,
        reader: MetadataRecordBatchReader,
        writer: FlightMetadataWriter,
    ) -> None: ...
    def do_get(self, context: ServerCallContext, ticket: Ticket) -> FlightDataStream: ...
    def do_exchange(
        self,
        context: ServerCallContext,
        descriptor: FlightDescriptor,
        reader: MetadataRecordBatchReader,
        writer: MetadataRecordBatchWriter,
    ) -> None: ...
    def list_actions(self, context: ServerCallContext) -> Iterable[Action]: ...
    def do_action(self, context: ServerCallContext, action: Action) -> Iterable[bytes]: ...
    def serve(self) -> None: ...
    def run(self) -> None: ...
    def shutdown(self) -> None: ...
    def wait(self) -> None: ...
    def __enter__(self) -> Self: ...
    def __exit__(self, exc_type, exc_value, traceback): ...

def connect(
    location: str | tuple[str, int] | Location,
    *,
    tls_root_certs: str | None = None,
    cert_chain: str | None = None,
    private_key: str | None = None,
    override_hostname: str | None = None,
    middleware: list[ClientMiddlewareFactory] | None = None,
    write_size_limit_bytes: int | None = None,
    disable_server_verification: bool = False,
    generic_options: list[tuple[str, int | str]] | None = None,
) -> FlightClient: ...



import sys

from collections.abc import Callable
from io import IOBase

from _typeshed import StrPath

if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self
if sys.version_info >= (3, 10):
    from typing import TypeAlias
else:
    from typing_extensions import TypeAlias

from typing import Any, Literal, SupportsIndex, overload

from pyarrow._stubs_typing import Compression, SupportPyBuffer
from pyarrow.lib import MemoryPool, _Weakrefable

from .device import Device, DeviceAllocationType, MemoryManager
from .types import KeyValueMetadata

def have_libhdfs() -> bool: ...
def io_thread_count() -> int: ...
def set_io_thread_count(count: int) -> None: ...

Mode: TypeAlias = Literal["rb", "wb", "rb+", "ab"]

class NativeFile(_Weakrefable):
    _default_chunk_size: int

    def __enter__(self) -> Self: ...
    def __exit__(self, *args) -> None: ...
    @property
    def mode(self) -> Mode: ...
    def readable(self) -> bool: ...
    def seekable(self) -> bool: ...
    def isatty(self) -> bool: ...
    def fileno(self) -> int: ...
    @property
    def closed(self) -> bool: ...
    def close(self) -> None: ...
    def size(self) -> int: ...
    def metadata(self) -> KeyValueMetadata: ...
    def tell(self) -> int: ...
    def seek(self, position: int, whence: int = 0) -> int: ...
    def flush(self) -> None: ...
    def write(self, data: bytes | SupportPyBuffer) -> int: ...
    def read(
        self,
    ) -> bytes: ...
    def get_stream(self, file_offset: int, nbytes: int) -> Self: ...
    def read_at(self) -> bytes: ...
    def read1(self) -> bytes: ...
    def readall(self) -> bytes: ...
    def readinto(self, b: SupportPyBuffer) -> int: ...
    def readline(self, size: int | None = None) -> bytes: ...
    def readlines(self, hint: int | None = None) -> list[bytes]: ...
    def __iter__(self) -> Self: ...
    def __next__(self) -> bytes: ...
    def read_buffer(self, nbytes: int | None = None) -> Buffer: ...
    def truncate(self) -> None: ...
    def writelines(self, lines: list[bytes]): ...
    def download(
        self, stream_or_path: StrPath | IOBase, buffer_size: int | None = None
    ) -> None: ...
    def upload(self, stream: IOBase, buffer_size: int | None) -> None: ...

# ----------------------------------------------------------------------
# Python file-like objects

class PythonFile(NativeFile):
    def __init__(self, handle: IOBase, mode: Literal["r", "w"] | None = None) -> None: ...
    def truncate(self, pos: int | None = None) -> None: ...

class MemoryMappedFile(NativeFile):
    @classmethod
    def create(cls, path: str, size: int) -> Self: ...
    def _open(self, path: str, mode: Literal["r", "rb", "w", "wb", "r+", "r+b", "rb+"] = "r"): ...
    def resize(self, new_size: int) -> None: ...

def memory_map(
    path: str, mode: Literal["r", "rb", "w", "wb", "r+", "r+b", "rb+"] = "r"
) -> MemoryMappedFile: ...

create_memory_map = MemoryMappedFile.create

class OSFile(NativeFile):
    def __init__(
        self,
        path: str,
        mode: Literal["r", "rb", "w", "wb", "a", "ab"],
        memory_pool: MemoryPool | None = None,
    ) -> None: ...

class FixedSizeBufferWriter(NativeFile):
    def __init__(self, buffer: Buffer) -> None: ...
    def set_memcopy_threads(self, num_threads: int) -> None: ...
    def set_memcopy_blocksize(self, blocksize: int) -> None: ...
    def set_memcopy_threshold(self, threshold: int) -> None: ...

# ----------------------------------------------------------------------
# Arrow buffers

class Buffer(_Weakrefable):
    def __len__(self) -> int: ...
    def _assert_cpu(self) -> None: ...
    @property
    def size(self) -> int: ...
    @property
    def address(self) -> int: ...
    def hex(self) -> bytes: ...
    @property
    def is_mutable(self) -> bool: ...
    @property
    def is_cpu(self) -> bool: ...
    @property
    def device(self) -> Device: ...
    @property
    def memory_manager(self) -> MemoryManager: ...
    @property
    def device_type(self) -> DeviceAllocationType: ...
    @property
    def parent(self) -> Buffer | None: ...
    @overload
    def __getitem__(self, key: slice) -> Self: ...
    @overload
    def __getitem__(self, key: int) -> int: ...
    def slice(self, offset: int = 0, length: int | None = None) -> Self: ...
    def equals(self, other: Self) -> bool: ...
    def __reduce_ex__(self, protocol: SupportsIndex) -> str | tuple[Any, ...]: ...
    def to_pybytes(self) -> bytes: ...
    def __buffer__(self, flags: int, /) -> memoryview: ...

class ResizableBuffer(Buffer):
    def resize(self, new_size: int, shrink_to_fit: bool = False) -> None: ...

@overload
def allocate_buffer(size: int, memory_pool: MemoryPool | None = None) -> Buffer: ...
@overload
def allocate_buffer(
    size: int, memory_pool: MemoryPool | None, resizable: Literal[False]
) -> Buffer: ...
@overload
def allocate_buffer(
    size: int, memory_pool: MemoryPool | None, resizable: Literal[True]
) -> ResizableBuffer: ...

# ----------------------------------------------------------------------
# Arrow Stream
class BufferOutputStream(NativeFile):
    def __init__(self, memory_pool: MemoryPool | None = None) -> None: ...
    def getvalue(self) -> Buffer: ...

class MockOutputStream(NativeFile): ...

class BufferReader(NativeFile):
    def __init__(self, obj) -> None: ...

class CompressedInputStream(NativeFile):
    def __init__(
        self,
        stream: StrPath | NativeFile | IOBase,
        compression: Literal["bz2", "brotli", "gzip", "lz4", "zstd"],
    ) -> None: ...

class CompressedOutputStream(NativeFile):
    def __init__(
        self,
        stream: StrPath | NativeFile | IOBase,
        compression: Literal["bz2", "brotli", "gzip", "lz4", "zstd"],
    ) -> None: ...

class BufferedInputStream(NativeFile):
    def __init__(
        self, stream: NativeFile, buffer_size: int, memory_pool: MemoryPool | None = None
    ) -> None: ...
    def detach(self) -> NativeFile: ...

class BufferedOutputStream(NativeFile):
    def __init__(
        self, stream: NativeFile, buffer_size: int, memory_pool: MemoryPool | None = None
    ) -> None: ...
    def detach(self) -> NativeFile: ...

class TransformInputStream(NativeFile):
    def __init__(self, stream: NativeFile, transform_func: Callable[[Buffer], Any]) -> None: ...

class Transcoder:
    def __init__(self, decoder, encoder) -> None: ...
    def __call__(self, buf: Buffer): ...

def transcoding_input_stream(
    stream: NativeFile, src_encoding: str, dest_encoding: str
) -> TransformInputStream: ...
def py_buffer(obj: SupportPyBuffer) -> Buffer: ...
def foreign_buffer(address: int, size: int, base: Any | None = None) -> Buffer: ...
def as_buffer(o: Buffer | SupportPyBuffer) -> Buffer: ...

# ---------------------------------------------------------------------

class CacheOptions(_Weakrefable):
    hole_size_limit: int
    range_size_limit: int
    lazy: bool
    prefetch_limit: int
    def __init__(
        self,
        *,
        hole_size_limit: int | None = None,
        range_size_limit: int | None = None,
        lazy: bool = True,
        prefetch_limit: int = 0,
    ) -> None: ...
    @classmethod
    def from_network_metrics(
        cls,
        time_to_first_byte_millis: int,
        transfer_bandwidth_mib_per_sec: int,
        ideal_bandwidth_utilization_frac: float = 0.9,
        max_ideal_request_size_mib: int = 64,
    ) -> Self: ...

class Codec(_Weakrefable):
    def __init__(self, compression: Compression, compression_level: int | None = None) -> None: ...
    @classmethod
    def detect(cls, path: StrPath) -> Self: ...
    @staticmethod
    def is_available(compression: Compression) -> bool: ...
    @staticmethod
    def supports_compression_level(compression: Compression) -> int: ...
    @staticmethod
    def default_compression_level(compression: Compression) -> int: ...
    @staticmethod
    def minimum_compression_level(compression: Compression) -> int: ...
    @staticmethod
    def maximum_compression_level(compression: Compression) -> int: ...
    @property
    def name(self) -> Compression: ...
    @property
    def compression_level(self) -> int: ...
    @overload
    def compress(
        self,
        buf: Buffer | bytes | SupportPyBuffer,
        *,
        memory_pool: MemoryPool | None = None,
    ) -> Buffer: ...
    @overload
    def compress(
        self,
        buf: Buffer | bytes | SupportPyBuffer,
        *,
        asbytes: Literal[False],
        memory_pool: MemoryPool | None = None,
    ) -> Buffer: ...
    @overload
    def compress(
        self,
        buf: Buffer | bytes | SupportPyBuffer,
        *,
        asbytes: Literal[True],
        memory_pool: MemoryPool | None = None,
    ) -> bytes: ...
    @overload
    def decompress(
        self,
        buf: Buffer | bytes | SupportPyBuffer,
        decompressed_size: int | None = None,
        *,
        memory_pool: MemoryPool | None = None,
    ) -> Buffer: ...
    @overload
    def decompress(
        self,
        buf: Buffer | bytes | SupportPyBuffer,
        decompressed_size: int | None = None,
        *,
        asbytes: Literal[False],
        memory_pool: MemoryPool | None = None,
    ) -> Buffer: ...
    @overload
    def decompress(
        self,
        buf: Buffer | bytes | SupportPyBuffer,
        decompressed_size: int | None = None,
        *,
        asbytes: Literal[True],
        memory_pool: MemoryPool | None = None,
    ) -> bytes: ...

@overload
def compress(
    buf: Buffer | bytes | SupportPyBuffer,
    codec: Compression = "lz4",
    *,
    memory_pool: MemoryPool | None = None,
) -> Buffer: ...
@overload
def compress(
    buf: Buffer | bytes | SupportPyBuffer,
    codec: Compression = "lz4",
    *,
    asbytes: Literal[False],
    memory_pool: MemoryPool | None = None,
) -> Buffer: ...
@overload
def compress(
    buf: Buffer | bytes | SupportPyBuffer,
    codec: Compression = "lz4",
    *,
    asbytes: Literal[True],
    memory_pool: MemoryPool | None = None,
) -> bytes: ...
@overload
def decompress(
    buf: Buffer | bytes | SupportPyBuffer,
    decompressed_size: int | None = None,
    codec: Compression = "lz4",
    *,
    memory_pool: MemoryPool | None = None,
) -> Buffer: ...
@overload
def decompress(
    buf: Buffer | bytes | SupportPyBuffer,
    decompressed_size: int | None = None,
    codec: Compression = "lz4",
    *,
    asbytes: Literal[False],
    memory_pool: MemoryPool | None = None,
) -> Buffer: ...
@overload
def decompress(
    buf: Buffer | bytes | SupportPyBuffer,
    decompressed_size: int | None = None,
    codec: Compression = "lz4",
    *,
    asbytes: Literal[True],
    memory_pool: MemoryPool | None = None,
) -> bytes: ...
def input_stream(
    source: StrPath | Buffer | IOBase,
    compression: Literal["detect", "bz2", "brotli", "gzip", "lz4", "zstd"] = "detect",
    buffer_size: int | None = None,
) -> BufferReader: ...
def output_stream(
    source: StrPath | Buffer | IOBase,
    compression: Literal["detect", "bz2", "brotli", "gzip", "lz4", "zstd"] = "detect",
    buffer_size: int | None = None,
) -> NativeFile: ...

__all__ = [
    "have_libhdfs",
    "io_thread_count",
    "set_io_thread_count",
    "NativeFile",
    "PythonFile",
    "MemoryMappedFile",
    "memory_map",
    "create_memory_map",
    "OSFile",
    "FixedSizeBufferWriter",
    "Buffer",
    "ResizableBuffer",
    "allocate_buffer",
    "BufferOutputStream",
    "MockOutputStream",
    "BufferReader",
    "CompressedInputStream",
    "CompressedOutputStream",
    "BufferedInputStream",
    "BufferedOutputStream",
    "TransformInputStream",
    "Transcoder",
    "transcoding_input_stream",
    "py_buffer",
    "foreign_buffer",
    "as_buffer",
    "CacheOptions",
    "Codec",
    "compress",
    "decompress",
    "input_stream",
    "output_stream",
]



from typing import Any, TypedDict, TypeVar

import pandas as pd

from pandas import DatetimeTZDtype

from .lib import Array, DataType, Schema, Table

_T = TypeVar("_T")

def get_logical_type_map() -> dict[int, str]: ...
def get_logical_type(arrow_type: DataType) -> str: ...
def get_logical_type_from_numpy(pandas_collection) -> str: ...
def get_extension_dtype_info(column) -> tuple[str, dict[str, Any]]: ...

class _ColumnMetadata(TypedDict):
    name: str
    field_name: str
    pandas_type: int
    numpy_type: str
    metadata: dict | None

def get_column_metadata(
    column: pd.Series | pd.Index, name: str, arrow_type: DataType, field_name: str
) -> _ColumnMetadata: ...
def construct_metadata(
    columns_to_convert: list[pd.Series],
    df: pd.DataFrame,
    column_names: list[str],
    index_levels: list[pd.Index],
    index_descriptors: list[dict],
    preserve_index: bool,
    types: list[DataType],
) -> dict[bytes, bytes]: ...
def dataframe_to_types(
    df: pd.DataFrame, preserve_index: bool | None, columns: list[str] | None = None
) -> tuple[list[str], list[DataType], dict[bytes, bytes]]: ...
def dataframe_to_arrays(
    df: pd.DataFrame,
    schema: Schema,
    preserve_index: bool | None,
    nthreads: int = 1,
    columns: list[str] | None = None,
    safe: bool = True,
) -> tuple[Array, Schema, int]: ...
def get_datetimetz_type(values: _T, dtype, type_) -> tuple[_T, DataType]: ...
def make_datetimetz(unit: str, tz: str) -> DatetimeTZDtype: ...
def table_to_dataframe(
    options, table: Table, categories=None, ignore_metadata: bool = False, types_mapper=None
) -> pd.DataFrame: ...
def make_tz_aware(series: pd.Series, tz: str) -> pd.Series: ...



from typing import Any, Collection, Literal, Protocol, TypeAlias, TypeVar

import numpy as np

from numpy.typing import NDArray

from .__lib_pxi.array import BooleanArray, IntegerArray

ArrayLike: TypeAlias = Any
Order: TypeAlias = Literal["ascending", "descending"]
JoinType: TypeAlias = Literal[
    "left semi",
    "right semi",
    "left anti",
    "right anti",
    "inner",
    "left outer",
    "right outer",
    "full outer",
]
Compression: TypeAlias = Literal[
    "gzip", "bz2", "brotli", "lz4", "lz4_frame", "lz4_raw", "zstd", "snappy"
]
NullEncoding: TypeAlias = Literal["mask", "encode"]
NullSelectionBehavior: TypeAlias = Literal["drop", "emit_null"]
Mask: TypeAlias = list[bool | None] | NDArray[np.bool_] | BooleanArray
Indices: TypeAlias = list[int] | NDArray[np.integer] | IntegerArray

_T = TypeVar("_T")
SingleOrList: TypeAlias = list[_T] | _T

class SupportEq(Protocol):
    def __eq__(self, other) -> bool: ...

class SupportLt(Protocol):
    def __lt__(self, other) -> bool: ...

class SupportGt(Protocol):
    def __gt__(self, other) -> bool: ...

class SupportLe(Protocol):
    def __le__(self, other) -> bool: ...

class SupportGe(Protocol):
    def __ge__(self, other) -> bool: ...

FilterTuple: TypeAlias = (
    tuple[str, Literal["=", "==", "!="], SupportEq]
    | tuple[str, Literal["<"], SupportLt]
    | tuple[str, Literal[">"], SupportGt]
    | tuple[str, Literal["<="], SupportLe]
    | tuple[str, Literal[">="], SupportGe]
    | tuple[str, Literal["in", "not in"], Collection]
)

class Buffer(Protocol):
    def __buffer__(self, flags: int, /) -> memoryview: ...

class SupportPyBuffer(Protocol):
    def __buffer__(self, flags: int, /) -> memoryview: ...

class SupportArrowStream(Protocol):
    def __arrow_c_stream__(self, requested_schema=None) -> Any: ...

class SupportArrowArray(Protocol):
    def __arrow_c_array__(self, requested_schema=None) -> Any: ...

class SupportArrowDeviceArray(Protocol):
    def __arrow_c_device_array__(self, requested_schema=None, **kwargs) -> Any: ...

class SupportArrowSchema(Protocol):
    def __arrow_c_schema(self) -> Any: ...



import enum

from pyarrow.lib import _Weakrefable

class DeviceAllocationType(enum.Flag):
    CPU = enum.auto()
    CUDA = enum.auto()
    CUDA_HOST = enum.auto()
    OPENCL = enum.auto()
    VULKAN = enum.auto()
    METAL = enum.auto()
    VPI = enum.auto()
    ROCM = enum.auto()
    ROCM_HOST = enum.auto()
    EXT_DEV = enum.auto()
    CUDA_MANAGED = enum.auto()
    ONEAPI = enum.auto()
    WEBGPU = enum.auto()
    HEXAGON = enum.auto()

class Device(_Weakrefable):
    @property
    def type_name(self) -> str: ...
    @property
    def device_id(self) -> int: ...
    @property
    def is_cpu(self) -> bool: ...
    @property
    def device_type(self) -> DeviceAllocationType: ...

class MemoryManager(_Weakrefable):
    @property
    def device(self) -> Device: ...
    @property
    def is_cpu(self) -> bool: ...

def default_cpu_memory_manager() -> MemoryManager: ...

__all__ = ["DeviceAllocationType", "Device", "MemoryManager", "default_cpu_memory_manager"]



from pyarrow._parquet_encryption import (
    CryptoFactory,
    DecryptionConfiguration,
    EncryptionConfiguration,
    KmsClient,
    KmsConnectionConfig,
)

__all__ = [
    "CryptoFactory",
    "DecryptionConfiguration",
    "EncryptionConfiguration",
    "KmsClient",
    "KmsConnectionConfig",
]
